# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import base64
import inspect
import json
import math
import os
import random
import types
import warnings
from typing import List, Union, Dict

import dill
import numpy as np
import pandas as pd
from featureform import metadata
from featureform.proto import serving_pb2
from featureform.proto import serving_pb2_grpc
from featureform.providers import get_provider, Scalar, VectorType
from pandas.core.generic import NDFrame
from pandasql import sqldf

from . import progress_bar
from .register import FeatureColumnResource

from .constants import NO_RECORD_LIMIT
from .enums import FileFormat, ScalarType
from .file_utils import absolute_file_paths
from .local_cache import LocalCache
from .local_utils import (
    get_sql_transformation_sources,
    feature_df_with_entity,
    feature_df_from_csv,
    label_df_from_csv,
    merge_feature_into_ts,
)
from .resources import Model, SourceType, ComputationMode
from .sqlite_metadata import SQLiteMetadata
from .tls import insecure_channel, secure_channel
from .version import check_up_to_date


def check_feature_type(features):
    checked_features = []
    for feature in features:
        if isinstance(feature, tuple):
            checked_features.append(feature)
        elif isinstance(feature, str):
            # TODO: Need to identify how to pull the run id
            checked_features.append((feature, "default"))
        elif isinstance(feature, FeatureColumnResource):
            checked_features.append(feature.name_variant())
    return checked_features


class ServingClient:
    """
    The serving client is used to retrieve training sets and features for training and serving purposes.
    **Using the Serving Client:**
    ``` py
    import featureform as ff
    from featureform import Client
    client = Client()
    # example:
    dataset = client.training_set("fraud_training", "quickstart")
    training_dataset = dataset.repeat(10).shuffle(1000).batch(8)
    for feature_batch in training_dataset:
        # Train model
    ```
    """

    def __init__(self, host=None, local=False, insecure=False, cert_path=None):
        # This line ensures that the warning is only raised if ServingClient is instantiated directly
        # TODO: Remove this check once ServingClient is deprecated
        is_instantiated_directed = inspect.stack()[1].function != "__init__"
        if is_instantiated_directed:
            warnings.warn(
                "ServingClient is deprecated and will be removed in future versions; use Client instead.",
                PendingDeprecationWarning,
            )
        """
        Args:
            host (str): The hostname of the Featureform instance. Exclude if using Localmode.
            local (bool): True if using Localmode.
            insecure (bool): True if connecting to an insecure Featureform endpoint. False if using a self-signed or public TLS certificate
            cert_path (str): The path to a public certificate if using a self-signed certificate.
        """
        if local and host:
            raise ValueError("Host and local cannot both be set")
        if local:
            self.impl = LocalClientImpl()
        else:
            self.impl = HostedClientImpl(host, insecure, cert_path)

    def training_set(
        self,
        name,
        variant="",
        include_label_timestamp=False,
        model: Union[str, Model] = None,
    ):
        """Return an iterator that iterates through the specified training set.

        **Examples**:
        ``` py
            client = ff.Client()
            dataset = client.training_set("fraud_training", "quickstart")
            training_dataset = dataset.repeat(10).shuffle(1000).batch(8)
            for feature_batch in training_dataset:
                # Train model
        ```

        Args:
            name (str): Name of training set to be retrieved
            variant (str): Variant of training set to be retrieved

        Returns:
            training_set (Dataset): A training set iterator
        """
        return self.impl.training_set(name, variant, include_label_timestamp, model)

    def features(
        self, features, entities, model: Union[str, Model] = None, params: list = None
    ):
        """Returns the feature values for the specified entities.

        **Examples**:
        ``` py
            client = ff.Client()
            fpf = client.features([("avg_transactions", "quickstart")], {"user": "C1410926"})
            # Run features through model
        ```

        Args:
            features (list[(str, str)], list[str]): List of Name Variant Tuples
            entities (dict): Dictionary of entity name/value pairs

        Returns:
            features (numpy.Array): An Numpy array of feature values in the order given by the inputs
        """
        features = check_feature_type(features)
        return self.impl.features(features, entities, model, params)

    def close(self):
        """Closes the connection to the Featureform instance."""
        self.impl.close()


class HostedClientImpl:
    def __init__(self, host=None, insecure=False, cert_path=None):
        host = host or os.getenv("FEATUREFORM_HOST")
        if host is None:
            raise ValueError(
                "If not in local mode then `host` must be passed or the environment"
                " variable FEATUREFORM_HOST must be set."
            )
        check_up_to_date(False, "serving")
        self._channel = self._create_channel(host, insecure, cert_path)
        self._stub = serving_pb2_grpc.FeatureStub(self._channel)

    def _create_channel(self, host, insecure, cert_path):
        if insecure:
            return insecure_channel(host)
        else:
            return secure_channel(host, cert_path)

    def training_set(
        self, name, variation, include_label_timestamp, model: Union[str, Model] = None
    ):
        return Dataset(self._stub).from_stub(name, variation, model)

    def features(
        self, features, entities, model: Union[str, Model] = None, params: list = None
    ):
        req = serving_pb2.FeatureServeRequest()
        for name, value in entities.items():
            entity_proto = req.entities.add()
            entity_proto.name = name
            entity_proto.value = value
        for name, variation in features:
            feature_id = req.features.add()
            feature_id.name = name
            feature_id.version = variation
        if model is not None:
            req.model.name = model if isinstance(model, str) else model.name
        resp = self._stub.FeatureServe(req)

        feature_values = []
        for val in resp.values:
            parsed_value = parse_proto_value(val)
            value_type = type(parsed_value)

            # Ondemand features are returned as a byte array
            # which holds the pickled function
            if value_type == bytes:
                code = dill.loads(bytearray(parsed_value))
                func = types.FunctionType(code, globals(), "transformation")
                parsed_value = func(self, params, entities)
            # Vector features are returned as a Vector32 proto due
            # to the inability to use the `repeated` keyword in
            # in a `oneof` field
            elif value_type == serving_pb2.Vector32:
                parsed_value = parsed_value.value

            feature_values.append(parsed_value)

        return feature_values

    def _get_source_as_df(self, name, variant, limit):
        columns = self._get_source_columns(name, variant)
        data = self._get_source_data(name, variant, limit)
        return pd.DataFrame(data=data, columns=columns)

    def _get_source_data(self, name, variant, limit):
        id = serving_pb2.SourceID(name=name, version=variant)
        req = serving_pb2.SourceDataRequest(id=id, limit=limit)
        resp = self._stub.SourceData(req)
        data = []
        for rows in resp:
            row = [getattr(r, r.WhichOneof("value")) for r in rows.rows]
            data.append(row)
        return data

    def _get_source_columns(self, name, variant):
        id = serving_pb2.SourceID(name=name, version=variant)
        req = serving_pb2.SourceDataRequest(id=id)
        resp = self._stub.SourceColumns(req)
        return resp.columns

    def nearest(self, name, variant, vector, k):
        id = serving_pb2.FeatureID(name=name, version=variant)
        vec = serving_pb2.Vector32(value=vector)
        req = serving_pb2.NearestRequest(id=id, vector=vec, k=k)
        resp = self._stub.Nearest(req)
        return resp.entities

    def close(self):
        self._channel.close()


class LocalClientImpl:
    def __init__(self):
        self.db = SQLiteMetadata()
        self.local_cache = LocalCache()
        check_up_to_date(True, "serving")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.db.close()

    def get_training_set_dataframe(
        self, label, label_df, training_set_name, training_set_variant
    ) -> NDFrame:
        def get() -> pd.DataFrame:
            feature_columns = []

            # We will build the training set DF by merging each feature one by one into it.
            training_set_df = label_df
            features = self.db.get_training_set_features(
                training_set_name, training_set_variant
            )
            for feature in features:
                feature_variant = self.db.get_feature_variant(
                    feature["feature_name"], feature["feature_variant"]
                )
                feature_df = self.get_feature_dataframe(feature_variant)
                training_set_df = merge_feature_into_ts(
                    feature_variant, label, feature_df, training_set_df
                )
                feature_columns.append(
                    f"{feature['feature_name']}.{feature['feature_variant']}"
                )

            lag_features = self.db.get_training_set_lag_features(
                training_set_name, training_set_variant
            )
            if len(lag_features) > 0:
                timestamp_column = label["source_timestamp"]
                entity_column = label["source_entity"]
                label_column = "label"
                lag_sql_query = self.get_lag_features_sql_query(
                    lag_features,
                    feature_columns,
                    entity_column,
                    label_column,
                    timestamp_column,
                )

                globals()["source_0"] = training_set_df
                mysql = lambda q: sqldf(q, globals())
                training_set_df = mysql(lag_sql_query)

            return training_set_df

        return self.local_cache.get_or_put_training_set(
            training_set_name=training_set_name,
            training_set_variant=training_set_variant,
            func=get,
        )

    def training_set(
        self,
        training_set_name,
        training_set_variant,
        include_label_timestamp,
        model: Union[str, Model] = None,
    ):
        training_set = self.db.get_training_set_variant(
            training_set_name, training_set_variant
        )

        label = self.db.get_label_variant(
            training_set["label_name"], training_set["label_variant"]
        )
        label_df = self.get_label_dataframe(label)

        if model is not None:
            self._register_model(
                model,
                look_up_table="model_training_sets",
                association_name=training_set_name,
                association_variant=training_set_variant,
            )

        training_set_df = self.get_training_set_dataframe(
            label, label_df, training_set_name, training_set_variant
        )

        return self.convert_ts_df_to_dataset(
            label, training_set_df, include_label_timestamp
        )

    def get_lag_features_sql_query(
        self, lag_features, feature_columns, entity, label, ts
    ):
        """
        Returns the SQL query to compute the lag features.
        Input:
        - lag_features: dict

        Returns:
        - sql_query: string
        """

        if len(lag_features) == 0:
            return "SELECT * FROM source_0"

        features = ""
        for f in feature_columns:
            features = f'{features}, "{f}"' if features != "" else f'"{f}"'

        MAIN_SELECT = f"SELECT {entity}, {ts}, {features}"
        SUBQUERY = "FROM (SELECT * FROM (SELECT *, row_number FROM ("
        INNER_QUERY = "FROM (( SELECT * FROM source_0 )"
        CLOSING_QUERY = f") tt ) WHERE row_number=1 ORDER BY {ts} ASC ))"

        lag_columns = []
        lag_timestamps = []

        LAG_QUERIES = ""
        for i, lag_feature in enumerate(lag_features):
            feature_column_name = (
                f"{lag_feature['feature_name']}.{lag_feature['feature_variant']}"
            )
            lag_feature_column_name = lag_feature["feature_new_name"]
            lag = lag_feature["feature_lag"]

            lag_query = f"""
                         LEFT OUTER JOIN ( SELECT * FROM ( SELECT {entity} AS t{i}_entity, \"{feature_column_name}\" AS \"{lag_feature_column_name}\", {ts} AS t{i}_ts
                         FROM source_0) 
                         ORDER BY t{i}_ts ASC) t{i}
                         ON (t{i}_entity = {entity} AND DATETIME(t{i}_ts, \"+{lag} seconds\") <= {ts})
                        """

            LAG_QUERIES = (
                f"{LAG_QUERIES} {lag_query}" if LAG_QUERIES != "" else lag_query
            )
            MAIN_SELECT = f'{MAIN_SELECT}, "{lag_feature_column_name}"'

            lag_columns.append(f'"{lag_feature_column_name}"')
            lag_timestamps.append(f"t{i}_ts")

        lag_columns = ", ".join(lag_columns)
        lag_timestamps = " DESC, ".join(lag_timestamps)

        INNER_SELECT = f"SELECT {entity}, {ts}, {label}, {features}, {lag_columns}, ROW_NUMBER() over (PARTITION BY {entity}, {label}, {ts} ORDER BY {ts} DESC, {lag_timestamps} DESC) as row_number"
        FULL_QUERY = f"{MAIN_SELECT}, {label} {SUBQUERY} {INNER_SELECT} {INNER_QUERY} {LAG_QUERIES} {CLOSING_QUERY}"

        return FULL_QUERY

    def get_input_df(self, source_name, source_variant):
        if (
            self.db.is_transformation(source_name, source_variant)
            == SourceType.PRIMARY_SOURCE
        ):
            source = self.db.get_source_variant(source_name, source_variant)
            file_path = source["definition"]
            if FileFormat.get_format(file_path) == FileFormat.CSV:
                df = pd.read_csv(file_path)
            elif FileFormat.get_format(file_path) == FileFormat.PARQUET:
                df = pd.read_parquet(file_path)
            else:
                raise ValueError(f"Unsupported file format for {file_path}")
            return df
        elif (
            self.db.is_transformation(source_name, source_variant)
            == SourceType.DIRECTORY
        ):
            source = self.db.get_source_variant(source_name, source_variant)
            directory = source["definition"]
            return self.read_directory(directory)

        else:
            df = self.process_transformation(source_name, source_variant)
        return df

    def read_directory(self, directory):
        if not os.path.isdir(directory):
            raise Exception(f"Path {directory} is not a directory")

        file_names = []
        file_body = []
        for absolute_fn, relative_fn in absolute_file_paths(directory):
            file_names.append(relative_fn)
            with open(absolute_fn, "r") as f:
                try:
                    file_body.append(f.read())
                except Exception as e:
                    raise Exception(
                        f"Cannot read file {absolute_fn}: {e}\nFiles must be text files"
                    )
        df = pd.DataFrame(data={"filename": file_names, "body": file_body})
        return df

    def sql_transformation(self, query):
        transformation_sources = get_sql_transformation_sources(query)
        for i, (source_name, source_variant) in enumerate(transformation_sources):
            # Creates a variable called dataframes_i which stores the corresponding df for each input
            df_variable = f"dataframes_{i}"
            # globals()[df_variable]:
            # 1. Converts a string "dataframes_i" to a variable name
            # 2. Assigns a global scope to the variable, to access it outside the loop
            globals()[df_variable] = self.get_input_df(source_name, source_variant)

            # using '+' signs for readability
            query_source_to_replace = "{{" + f"{source_name}.{source_variant}" + "}}"
            query = query.replace(query_source_to_replace, df_variable)

        return sqldf(query, globals())

    def process_transformation(self, name, variant):
        def get():
            source = self.db.get_source_variant(name, variant)
            if (
                self.db.is_transformation(name, variant)
                == SourceType.SQL_TRANSFORMATION.value
            ):
                query = source["definition"]
                new_data = self.sql_transformation(query)
            else:
                code = dill.loads(bytearray(source["definition"]))
                inputs = json.loads(source["inputs"])
                dataframes = []
                for input in inputs:
                    source_name, source_variant = (
                        input[0],
                        input[1],
                    )
                    dataframes.append(self.get_input_df(source_name, source_variant))
                func = types.FunctionType(code, globals(), "transformation")
                new_data = func(*dataframes)
                if new_data is None:
                    raise ValueError(
                        f"Transformation {name} ({variant}) returned None. Please return a dataframe."
                    )
            return new_data

        return self.local_cache.get_or_put(
            resource_type="transformation",
            resource_name=name,
            resource_variant=variant,
            source_name=name,
            source_variant=variant,
            func=get,
        )

    def get_label_dataframe(self, label) -> NDFrame:
        def get() -> pd.DataFrame:
            transform_type = self.db.is_transformation(
                label["source_name"], label["source_variant"]
            )
            if (
                transform_type == SourceType.SQL_TRANSFORMATION.value
                or transform_type == SourceType.DF_TRANSFORMATION.value
            ):
                label_df = self.label_df_from_transformation(label)
            else:
                label_source = self.db.get_source_variant(
                    label["source_name"], label["source_variant"]
                )
                label_df = label_df_from_csv(label, label_source["definition"])
            label_df.rename(columns={label["source_value"]: "label"}, inplace=True)
            return label_df

        try:
            return self.local_cache.get_or_put(
                resource_type="label",
                resource_name=label["name"],
                resource_variant=label["variant"],
                source_name=label["source_name"],
                source_variant=label["source_variant"],
                func=get,
            )
        except Exception as e:
            raise ValueError(
                f"Could not get source for label {label['name']} ({label['variant']}): {e}"
            )

    def label_df_from_transformation(self, label):
        df = self.process_transformation(label["source_name"], label["source_variant"])
        if label["source_timestamp"] != "":
            df = df[
                [
                    label["source_entity"],
                    label["source_value"],
                    label["source_timestamp"],
                ]
            ]
            df[label["source_timestamp"]] = pd.to_datetime(
                df[label["source_timestamp"]]
            )
        else:
            df = df[[label["source_entity"], label["source_value"]]]
        df.set_index(label["source_entity"])
        return df

    def get_feature_dataframe(self, feature) -> NDFrame:
        def get() -> pd.DataFrame:
            name_variant = feature["name"] + "." + feature["variant"]
            transform_type = self.db.is_transformation(
                feature["source_name"], feature["source_variant"]
            )
            if (
                transform_type == SourceType.SQL_TRANSFORMATION.value
                or transform_type == SourceType.DF_TRANSFORMATION.value
            ):
                feature_df = self.feature_df_from_transformation(feature)
            else:
                source = self.db.get_source_variant(
                    feature["source_name"], feature["source_variant"]
                )
                feature_df = feature_df_from_csv(feature, source["definition"])
            feature_df.set_index(feature["source_entity"])
            feature_df.rename(
                columns={feature["source_value"]: name_variant}, inplace=True
            )
            return feature_df

        return self.local_cache.get_or_put(
            resource_type="feature",
            resource_name=feature["name"],
            resource_variant=feature["variant"],
            source_name=feature["source_name"],
            source_variant=feature["source_variant"],
            func=get,
        )

    def feature_df_from_transformation(self, feature):
        df = self.process_transformation(
            feature["source_name"], feature["source_variant"]
        )
        if isinstance(df, pd.Series):
            df = df.to_frame()
            df.reset_index(inplace=True)
        if feature["source_timestamp"] != "" and feature["source_timestamp"] not in df:
            raise ValueError(
                f"Provided timestamp column '{feature['source_timestamp']}' for feature "
                f"'{feature['name']}:{feature['variant']}' not found in source '{feature['source_name']}:{feature['source_variant']}'; "
                f"either remove 'timestamp_column' from the feature registration or include it in the source."
            )
        elif feature["source_timestamp"] != "":
            df = df[
                [
                    feature["source_entity"],
                    feature["source_value"],
                    feature["source_timestamp"],
                ]
            ]
            df[feature["source_timestamp"]] = pd.to_datetime(
                df[feature["source_timestamp"]]
            )
        else:
            df = df[[feature["source_entity"], feature["source_value"]]]
        return df

    def features(
        self,
        feature_variant_list,
        entities: Dict,
        model: Union[str, Model] = None,
        params: list = None,
    ):
        if len(feature_variant_list) == 0:
            raise Exception("No features provided")

        self.entities = entities
        self.params = params if params else []

        self.__validate_entity_exists(entities, feature_variant_list)

        entity_name = list(entities.keys())[0] if len(entities) > 0 else ""
        entity_value = entities[entity_name] if len(entities) > 0 else ""
        features = self.add_features_to_list(
            feature_variant_list, entity_name, entity_value
        )
        # all_features_df = list_to_combined_df(all_features_list, entity_name)
        # features = get_features_for_entity(entity_name, entity_value, all_features_df)

        if model is not None:
            for feature_name, feature_variant in feature_variant_list:
                self._register_model(
                    model,
                    look_up_table="model_features",
                    association_name=feature_name,
                    association_variant=feature_variant,
                )

        return features

    def __validate_entity_exists(self, entities, feature_variant_list):
        # validate entities exists if any of the features are not ondemand
        if any(
            [
                self.db.get_feature_variant_mode(f_name, f_variant)
                != ComputationMode.CLIENT_COMPUTED
                for f_name, f_variant in feature_variant_list
            ]
        ):
            if len(entities) == 0:
                raise Exception("Entities are required for features (unless ondemand)")

    def calculate_ondemand_feature(self, f_name, f_variant):
        query = self.db.get_ondemand_feature_query(f_name, f_variant)
        base64_bytes = query.encode("ascii")
        query = base64.b64decode(base64_bytes)

        code = dill.loads(bytearray(query))
        func = types.FunctionType(code, globals(), "transformation")
        return func(self, self.params, self.entities)

    def add_features_to_list(self, feature_variant_list, entity_name, entity_value):
        feature_list = []

        for feature_variant in feature_variant_list:
            f_name = feature_variant[0]
            f_variant = feature_variant[1]
            f_mode = self.db.get_feature_variant_mode(f_name, f_variant)

            if f_mode == ComputationMode.CLIENT_COMPUTED:
                output_value = self.calculate_ondemand_feature(f_name, f_variant)
                feature_list.append(output_value)
            else:
                self.compute_feature(f_name, f_variant, entity_name)
                feature_df = self.get_feature_value(f_name, f_variant, entity_value)

                feature_list.append(feature_df)

        return np.array(feature_list)

    def compute_feature(self, f_name, f_variant, entity_name):
        feature = self.db.get_feature_variant(f_name, f_variant)
        source_name, source_variant = feature["source_name"], feature["source_variant"]

        source_files_from_db = self.db.get_source_files_for_resource(
            "transformation", source_name, source_variant
        )

        provider_obj = metadata.get_provider(feature["provider"])
        provider_type = provider_obj.function
        # This will be replaced to select the appropriate provider for each feature
        provider = get_provider(provider_type)(provider_obj.config)
        table_exists = provider.table_exists(f_name, f_variant)

        if (
            not any(
                self._file_has_changed(
                    source_file["updated_at"], source_file["file_path"]
                )
                for source_file in source_files_from_db
            )
            and len(source_files_from_db) > 0
            and table_exists
        ):
            return

        if feature["entity"] != entity_name:
            raise ValueError(
                f"Invalid entity {entity_name} for feature {source_name}-{source_variant}"
            )
        if (
            self.db.is_transformation(source_name, source_variant)
            != SourceType.PRIMARY_SOURCE.value
        ):
            feature_df = self.process_non_primary_df_transformation(
                feature, source_name, source_variant, entity_name
            )
        else:
            source = self.db.get_source_variant(source_name, source_variant)
            feature_df = feature_df_with_entity(
                source["definition"], entity_name, feature
            )

        if table_exists:
            table = provider.get_table(f_name, f_variant)
        else:
            if not feature["is_embedding"]:
                table = provider.create_table(
                    f_name, f_variant, Scalar(ScalarType(feature["data_type"]))
                )
            else:
                table = provider.create_index(
                    f_name,
                    f_variant,
                    VectorType(
                        ScalarType(feature["data_type"]), feature["dimension"], True
                    ),
                )

        total = len(feature_df)
        if provider_type == "LOCAL_ONLINE":
            table.set_batch(feature_df)
        else:
            for index, row in feature_df.iterrows():
                table.set(row[0], row[1])
                progress_bar(
                    total,
                    index,
                    prefix="Updating Feature Table:",
                    suffix="Complete",
                    length=50,
                )
        progress_bar(
            total, total, prefix="Updating Feature Table:", suffix="Complete", length=50
        )
        print("\n")

    @staticmethod
    def _file_has_changed(last_updated_at, file_path):
        """
        Currently using last updated at for determining if a file has changed. We can consider using the file hash
        if this becomes a performance issue.
        """
        os_last_updated = os.path.getmtime(file_path)
        return os_last_updated > float(last_updated_at)

    def get_feature_value(self, f_name, f_variant, entity_value):
        feature = self.db.get_feature_variant(f_name, f_variant)
        provider_obj = metadata.get_provider(feature["provider"])
        provider_type = provider_obj.function
        provider = get_provider(provider_type)(provider_obj.config)
        table = provider.get_table(f_name, f_variant)
        value = table.get(entity_value)

        return value

    def process_non_primary_df_transformation(
        self, feature, source_name, source_variant, entity_id
    ):
        name_variant = f"{feature['name']}.{feature['variant']}"
        feature_df = self.process_transformation(source_name, source_variant)
        if isinstance(feature_df, pd.Series):
            feature_df = feature_df.to_frame()
            feature_df.reset_index(inplace=True)
        if not feature["source_entity"] in feature_df.columns:
            raise ValueError(
                f"Could not set entity column. No column name {feature['source_entity']} exists in {source_name} ({source_variant})"
            )
        if not feature["source_value"] in feature_df.columns:
            raise ValueError(
                f"Could not access feature value column. No column name '{feature['source_value']}' exists in {source_name} ({source_variant})"
            )
        feature_df = feature_df[[feature["source_entity"], feature["source_value"]]]
        feature_df.rename(
            columns={
                feature["source_entity"]: entity_id,
                feature["source_value"]: name_variant,
            },
            inplace=True,
        )
        feature_df.drop_duplicates(subset=[entity_id], keep="last", inplace=True)
        feature_df.set_index(entity_id)
        return feature_df

    @staticmethod
    def convert_ts_df_to_dataset(label_row, trainingset_df, include_label_timestamp):
        if label_row["source_timestamp"] != "" and include_label_timestamp != True:
            trainingset_df.drop(columns=label_row["source_timestamp"], inplace=True)
        elif label_row["source_timestamp"] != "" and include_label_timestamp:
            source_timestamp_col = trainingset_df.pop(label_row["source_timestamp"])
            trainingset_df = trainingset_df.assign(label_timestamp=source_timestamp_col)
        trainingset_df.drop(columns=label_row["source_entity"], inplace=True)

        label_col = trainingset_df.pop("label")
        trainingset_df = trainingset_df.assign(label=label_col)
        return Dataset.from_dataframe(trainingset_df, include_label_timestamp)

    def _register_model(
        self,
        model: Union[str, Model],
        look_up_table: str,
        association_name: str,
        association_variant: str,
    ):
        name = model if isinstance(model, str) else model.name
        type = "Model" if isinstance(model, str) else model.type()

        self.db.insert("models", name, type)
        self.db.insert(look_up_table, name, association_name, association_variant)

    def _get_source_as_df(self, name, variant, limit):
        if limit == 0:
            raise ValueError("limit must be greater than 0")
        df = self.get_input_df(name, variant)
        if limit != NO_RECORD_LIMIT:
            return df[:limit]
        else:
            return df

    def nearest(self, name, variant, vector, k):
        feature = self.db.get_feature_variant(name, variant)
        self.compute_feature(name, variant, feature["entity"])
        provider_obj = metadata.get_provider(feature["provider"])
        provider_type = provider_obj.function
        provider = get_provider(provider_type)(provider_obj.config)

        if provider.table_exists(name, variant):
            table = provider.get_table(name, variant)
        else:
            raise ValueError(f"Table does not exist for feature {name} ({variant})")
        return table.nearest(name, variant, vector, k)

    def close(self):
        self.db.close()


class Stream:
    def __init__(self, stub, name, version, model: Union[str, Model] = None):
        req = serving_pb2.TrainingDataRequest()
        req.id.name = name
        req.id.version = version
        if model is not None:
            req.model.name = model if isinstance(model, str) else model.name
        self.name = name
        self.version = version
        self._stub = stub
        self._req = req
        self._iter = stub.TrainingData(req)

    def __iter__(self):
        return self

    def __next__(self):
        return Row(next(self._iter))

    def restart(self):
        self._iter = self._stub.TrainingData(self._req)


class LocalStream:
    def __init__(self, datalist, include_label_timestamp):
        self._datalist = datalist
        self._iter = iter(datalist)
        self._include_label_timestamp = include_label_timestamp

    def __iter__(self):
        return iter(self._iter)

    def __next__(self):
        return LocalRow(next(self._iter), self._include_label_timestamp)

    def restart(self):
        self._iter = iter(self._datalist)


class Repeat:
    def __init__(self, repeat_num, stream):
        self.repeat_num = repeat_num
        self._stream = stream

    def __iter__(self):
        return self

    def __next__(self):
        try:
            next_val = next(self._stream)
        except StopIteration:
            self.repeat_num -= 1
            if self.repeat_num >= 0:
                self._stream.restart()
                next_val = next(self._stream)
            else:
                raise

        return next_val


class Shuffle:
    def __init__(self, buffer_size, stream):
        self.buffer_size = buffer_size
        self._shuffled_data_list = []
        self._stream = stream
        self.__setup_buffer()

    def __setup_buffer(self):
        try:
            for _ in range(self.buffer_size):
                self._shuffled_data_list.append(next(self._stream))
        except StopIteration:
            pass

    def restart(self):
        self._stream.restart()
        self.__setup_buffer()

    def __iter__(self):
        return self

    def __next__(self):
        if len(self._shuffled_data_list) == 0:
            raise StopIteration
        random_index = random.randrange(len(self._shuffled_data_list))
        next_row = self._shuffled_data_list.pop(random_index)

        try:
            self._shuffled_data_list.append(next(self._stream))
        except StopIteration:
            pass

        return next_row


class Batch:
    def __init__(self, batch_size, stream):
        self.batch_size = batch_size
        self._stream = stream

    def restart(self):
        self._stream.restart()

    def __iter__(self):
        return self

    def __next__(self):
        rows = BatchRow()
        for _ in range(self.batch_size):
            try:
                next_row = next(self._stream)
                rows.append(next_row)
            except StopIteration:
                if len(rows) == 0:
                    raise
                return rows
        return rows


class Dataset:
    def __init__(self, stream, dataframe=None):
        """Repeats the Dataset for the specified number of times

        Args:
            stream (Iterator): An iterable object.

        Returns:
            self (Dataset): Returns a Dataset created from the iterable object.
        """
        self._stream = stream
        self._dataframe = dataframe

    def from_stub(self, name, version, model: Union[str, Model] = None):
        stream = Stream(self._stream, name, version, model)
        return Dataset(stream)

    def dataframe(self) -> pd.DataFrame:
        """Returns the training set as a pandas DataFrame

        **Examples**:
        ``` py
            client = Client()
            df = client.training_set("fraud_training", "v1").dataframe()
            print(df.head())
            # Output:
            #                feature_avg_transactions_default  label_fraudulent_default
            # 0                                25.0                       False
            # 1                             27999.0                       False
            # 2                               459.0                       False
            # 3                              2060.0                        True
            # 4                              1762.5                       False
        ```

        Returns:
            pandas.DataFrame: A pandas DataFrame containing the training set.
        """
        if self._dataframe is not None:
            return self._dataframe
        else:
            name = self._stream.name
            variant = self._stream.version
            stub = self._stream._stub
            id = serving_pb2.TrainingDataID(name=name, version=variant)
            req = serving_pb2.TrainingDataRequest(id=id)
            cols = stub.TrainingDataColumns(req)
            data = [r.to_dict(cols.features, cols.label) for r in self._stream]
            self._dataframe = pd.DataFrame(
                data=data, columns=[*cols.features, cols.label]
            )
            return self._dataframe

    def from_dataframe(dataframe, include_label_timestamp):
        stream = LocalStream(dataframe.values.tolist(), include_label_timestamp)
        return Dataset(stream, dataframe)

    def pandas(self):
        warnings.warn(
            "Dataset.pandas() is deprecated and will be removed in future versions; use Dataset.dataframe() instead.",
            PendingDeprecationWarning,
        )
        if self._dataframe is None:
            _ = self.dataframe()
        return self._dataframe

    def repeat(self, num):
        """Repeats the Dataset for the specified number of times

        **Examples**:
        ``` py
            client = ff.Client()
            dataset = client.training_set("fraud_training", "quickstart")
            training_dataset = dataset.repeat(10) # Repeats data 10 times
            for feature_batch in training_dataset:
                # Train model
        ```

        Args:
            num (int): The number of times the dataset will be repeated

        Returns:
            self (Dataset): Returns the current Dataset
        """
        if num <= 0:
            raise Exception("Must repeat 1 or more times")
        self._stream = Repeat(num, self._stream)
        if self._dataframe is not None:
            temp_df = self._dataframe
            for i in range(num):
                self._dataframe = self._dataframe.append(temp_df)
        return self

    def shuffle(self, buffer_size):
        """Swaps random rows within the Dataset.

        **Examples**:
        ``` py
            client = ff.Client()
            dataset = client.training_set("fraud_training", "quickstart")
            training_dataset = dataset.shuffle(100) # Swaps 100 Rows
            for feature_batch in training_dataset:
                # Train model
        ```

        Args:
            buffer_size (int): The number of Dataset rows to be randomly swapped

        Returns:
            self (Dataset): Returns the current Dataset
        """
        if buffer_size <= 0:
            raise Exception("Buffer size must be greater than or equal to 1")
        self._stream = Shuffle(buffer_size, self._stream)
        if self._dataframe is not None:
            self._dataframe = self._dataframe.sample(frac=1)
        return self

    def batch(self, batch_size):
        """Creates a batch row in the Dataset.

        **Examples**:
        ``` py
            client = ff.Client()
            dataset = client.training_set("fraud_training", "quickstart")
            training_dataset = dataset.batch(8) # Creates a batch of 8 Datasets for each row
            for feature_batch in training_dataset:
                # Train model
        ```

        Args:
            batch_size (int): The number of items to be added to each batch

        Returns:
            self (Dataset): Returns the current Dataset
        """
        if batch_size <= 0:
            raise Exception("Batch size must be greater than or equal to 1")
        self._stream = Batch(batch_size, self._stream)
        if self._dataframe is not None:
            self._dataframe = np.array_split(
                self._dataframe, math.ceil(len(self._dataframe) // batch_size)
            )
        return self

    def __iter__(self):
        return self

    def __next__(self):
        next_val = next(self._stream)
        return next_val


class Row:
    def __init__(self, proto_row):
        self._features = np.array(
            [parse_proto_value(feature) for feature in proto_row.features]
        )
        self._label = parse_proto_value(proto_row.label)
        self._row = np.append(self._features, self._label)

    def features(self):
        return [self._row[:-1]]

    def label(self):
        return [self._label]

    def to_numpy(self):
        return self._row

    def to_dict(self, feature_columns: List[str], label_column: str):
        row_dict = dict(zip(feature_columns, self._features))
        row_dict[label_column] = self._label
        return row_dict

    def __repr__(self):
        return "Features: {} , Label: {}".format(self.features(), self.label())


class LocalRow:
    def __init__(self, row_list, include_label_timestamp):
        """
        If include_label_timestamp is true then we want the label to equal to the
        last two columns in the list. Otherwise, just the label will be the last column only.
        """

        self._features = row_list[:-2] if include_label_timestamp else row_list[:-1]
        self._row = row_list
        self._label = row_list[-2:] if include_label_timestamp else row_list[-1]

    def features(self):
        return [self._features]

    def label(self):
        return [self._label]

    def to_numpy(self):
        return np.array(self._row)

    def __repr__(self):
        return "Features: {} , Label: {}".format(self.features(), self.label())


class BatchRow:
    def __init__(self, rows=None):
        self._features = []
        self._labels = []
        if rows is None:
            rows = []
        self._rows = rows
        for row in rows:
            self.append(row)

    def append(self, row):
        self._features.append(row.features()[0])
        self._labels.append(row.label()[0])
        self._rows.append(row)

    def features(self):
        return self._features

    def label(self):
        return self._labels

    def to_list(self):
        return self._rows

    def __len__(self):
        return len(self._rows)


def parse_proto_value(value):
    """parse_proto_value is used to parse the one of Value message"""
    return getattr(value, value.WhichOneof("value"))



import featureform as ff
import pandas as pd
import pytest
from featureform.enums import FileFormat


@pytest.mark.parametrize(
    "provider_source_fxt,is_local,is_insecure",
    [
        pytest.param(
            "local_provider_source",
            True,
            True,
            marks=pytest.mark.local,
        ),
        pytest.param(
            "hosted_sql_provider_and_source", False, False, marks=pytest.mark.hosted
        ),
        pytest.param(
            "hosted_sql_provider_and_source", False, True, marks=pytest.mark.docker
        ),
    ],
)
def test_dataframe_for_name_variant_args(
    provider_source_fxt, is_local, is_insecure, request
):
    custom_marks = [
        mark.name for mark in request.node.own_markers if mark.name != "parametrize"
    ]
    provider, source, inference_store = request.getfixturevalue(provider_source_fxt)(
        custom_marks
    )

    transformation = arrange_transformation(provider, is_local)

    client = ff.Client(local=is_local, insecure=is_insecure)
    # If we're running in a hosted context, `apply` needs to be synchronous
    # to ensure resources are ready to test.
    client.apply(asynchronous=is_local)

    if is_local:
        source_df = client.dataframe(source.name, source.variant)
    else:
        source_df = client.dataframe(*source.name_variant())
    transformation_df = client.dataframe(*transformation.name_variant())

    assert isinstance(source_df, pd.DataFrame) and isinstance(
        transformation_df, (pd.DataFrame, pd.Series)
    )

    if is_local:
        client.impl.db.close()  # TODO automatically do this


@pytest.mark.parametrize(
    "provider_source_fxt,is_local,is_insecure",
    [
        pytest.param(
            "local_provider_source",
            True,
            True,
            marks=pytest.mark.local,
        ),
        pytest.param(
            "hosted_sql_provider_and_source", False, False, marks=pytest.mark.hosted
        ),
        pytest.param(
            "hosted_sql_provider_and_source", False, True, marks=pytest.mark.docker
        ),
    ],
)
def test_dataframe_for_source_args(provider_source_fxt, is_local, is_insecure, request):
    custom_marks = [
        mark.name for mark in request.node.own_markers if mark.name != "parametrize"
    ]
    provider, source, inference_store = request.getfixturevalue(provider_source_fxt)(
        custom_marks
    )

    transformation = arrange_transformation(provider, is_local)

    client = ff.Client(local=is_local, insecure=is_insecure)
    # If we're running in a hosted context, `apply` needs to be synchronous
    # to ensure resources are ready to test.
    client.apply(asynchronous=is_local)

    source_df = client.dataframe(source)
    transformation_df = client.dataframe(transformation)

    assert isinstance(source_df, pd.DataFrame) and isinstance(
        transformation_df, (pd.DataFrame, pd.Series)
    )

    if is_local:
        client.impl.db.close()  # TODO automatically do this


# Ensures that the dataframe method raises an error when the variant is not specified and the source is a string
def test_dataframe_empty_variant(local_provider_source):
    provider, source, inference_store = local_provider_source("_empty_param")
    arrange_transformation(provider, "true")

    client = ff.Client(local=True, insecure=True)
    client.apply(asynchronous=True)

    with pytest.raises(ValueError) as e:
        client.dataframe(source.name)
    assert "variant must be specified if source is a string" in str(e.value)


@pytest.mark.parametrize(
    "provider_source_fxt,is_local,is_insecure",
    [
        pytest.param(
            "local_provider_source",
            True,
            True,
            marks=pytest.mark.local,
        ),
    ],
)
def test_dataframe_parquet(provider_source_fxt, is_local, is_insecure, request):
    custom_marks = [
        mark.name for mark in request.node.own_markers if mark.name != "parametrize"
    ]
    provider, source, inference_store = request.getfixturevalue(provider_source_fxt)(
        custom_marks, file_format=FileFormat.PARQUET.value
    )

    transformation = arrange_transformation(provider, is_local)

    client = ff.Client(local=is_local, insecure=is_insecure)
    client.apply(asynchronous=True)

    source_df = client.dataframe(source)
    transformation_df = client.dataframe(transformation)

    assert isinstance(source_df, pd.DataFrame) and isinstance(
        transformation_df, (pd.DataFrame, pd.Series)
    )

    if is_local:
        client.impl.db.close()  # TODO automatically do this


@pytest.fixture(autouse=True)
def before_and_after_each(setup_teardown):
    setup_teardown()
    yield
    setup_teardown()


def arrange_transformation(provider, is_local):
    if is_local:

        @provider.df_transformation(
            variant="quickstart", inputs=[("transactions", "quickstart")]
        )
        def average_user_transaction(transactions):
            return transactions.groupby("CustomerID")["TransactionAmount"].mean()

    else:

        @provider.sql_transformation(variant="quickstart")
        def average_user_transaction():
            return "SELECT customerid as user_id, avg(transactionamount) AS avg_transaction_amt FROM {{transactions.quickstart}} GROUP BY user_id"

    return average_user_transaction



from typing import Union, Optional
from datetime import datetime
from typing import Union, List, Optional

import pandas as pd
from featureform.proto import metadata_pb2 as pb
from google.protobuf.timestamp_pb2 import Timestamp
from typeguard import typechecked

from .constants import NO_RECORD_LIMIT
from .register import (
    ResourceClient,
    SourceRegistrar,
    LocalSource,
    SubscriptableTransformation,
    FeatureColumnResource,
)
from .serving import ServingClient


class Client(ResourceClient, ServingClient):
    """
    Client for interacting with Featureform APIs (resources and serving)

    **Using the Client:**
    ```py title="definitions.py"
    import featureform as ff
    from featureform import Client

    client = Client()

    # Example 1: Get a registered provider
    redis = client.get_provider("redis-quickstart")

    # Example 2: Compute a dataframe from a registered source
    transactions_df = client.dataframe("transactions", "quickstart")
    ```
    """

    def __init__(
        self, host=None, local=False, insecure=False, cert_path=None, dry_run=False
    ):
        if host is not None:
            self._validate_host(host)

        ResourceClient.__init__(
            self,
            host=host,
            local=local,
            insecure=insecure,
            cert_path=cert_path,
            dry_run=dry_run,
        )
        # Given both ResourceClient and ServingClient are instantiated together, if dry_run is True, then
        # the ServingClient cannot be instantiated due to a conflict the local and host arguments.
        if not dry_run:
            ServingClient.__init__(
                self, host=host, local=local, insecure=insecure, cert_path=cert_path
            )

    def dataframe(
        self,
        source: Union[SourceRegistrar, LocalSource, SubscriptableTransformation, str],
        variant: Optional[str] = None,
        limit=NO_RECORD_LIMIT,
        asynchronous=False,
        verbose=False,
    ):
        """
        Return a dataframe from a registered source or transformation

        **Example:**
        ```py title="definitions.py"
        transactions_df = client.dataframe("transactions", "quickstart")

        avg_user_transaction_df = transactions_df.groupby("CustomerID")["TransactionAmount"].mean()
        ```

        Args:
            source (Union[SourceRegistrar, LocalSource, SubscriptableTransformation, str]): The source or transformation to compute the dataframe from
            variant (str): The source variant; can't be None if source is a string
            limit (int): The maximum number of records to return; defaults to NO_RECORD_LIMIT
            asynchronous (bool): Flag to determine whether the client should wait for resources to be in either a READY or FAILED state before returning. Defaults to False to ensure that newly registered resources are in a READY state prior to serving them as dataframes.

        Returns:
            df (pandas.DataFrame): The dataframe computed from the source or transformation

        """
        self.apply(asynchronous=asynchronous, verbose=verbose)
        if isinstance(
            source, (SourceRegistrar, LocalSource, SubscriptableTransformation)
        ):
            name, variant = source.name_variant()
        elif isinstance(source, str):
            name = source
            if variant is None:
                raise ValueError("variant must be specified if source is a string")
            if variant == "":
                raise ValueError("variant cannot be an empty string")
        else:
            raise ValueError(
                f"source must be of type SourceRegistrar, LocalSource, SubscriptableTransformation or str, not {type(source)}"
            )
        return self.impl._get_source_as_df(name, variant, limit)

    def nearest(self, feature, vector, k):
        """
        Query the K nearest neighbors of a provider vector in the index of a registered feature variant

        **Example:**

        ```py title="definitions.py"
        # Get the 5 nearest neighbors of the vector [0.1, 0.2, 0.3] in the index of the feature "my_feature" with variant "my_variant"
        nearest_neighbors = client.nearest("my_feature", "my_variant", [0.1, 0.2, 0.3], 5)
        print(nearest_neighbors) # prints a list of entities (e.g. ["entity1", "entity2", "entity3", "entity4", "entity5"])
        ```

        Args:
            feature (Union[FeatureColumnResource, tuple(str, str)]): Feature object or tuple of Feature name and variant
            vector (List[float]): Query vector
            k (int): Number of nearest neighbors to return

        """
        if isinstance(feature, tuple):
            name, variant = feature
        elif isinstance(feature, FeatureColumnResource):
            name = feature.name
            variant = feature.variant
        else:
            raise Exception(
                f"the feature '{feature}' of type '{type(feature)}' is not support."
                "Feature must be a tuple of (name, variant) or a FeatureColumnResource"
            )

        if k < 1:
            raise ValueError(f"k must be a positive integer")
        return self.impl._nearest(name, variant, vector, k)

    @typechecked
    def write_feature(
        self,
        name_variant: tuple,
        entity: str,
        value: Union[str, int, float, bool, List[float]],
    ):
        name, variant = name_variant

        ts = Timestamp()
        # consider allowing user to pass in timestamp
        ts.GetCurrentTime()

        feature = pb.StreamingFeatureVariant(
            name=name,
            variant=variant,
            entity=entity,
            value=str(value),
            ts=ts,
        )

        features = iter([feature])
        # TODO: add retry logic
        self._stub.WriteFeatures(features)

    @typechecked
    def write_label(
        self,
        name_variant: tuple,
        entity: str,
        value: Union[str, int, float, bool, List[float]],
        timestamp: Optional[datetime] = None,
    ):
        name, variant = name_variant

        ts = Timestamp()
        if timestamp is None:
            ts.GetCurrentTime()
        else:
            ts.FromDatetime(timestamp)

        label = pb.StreamingLabelVariant(
            name=name,
            variant=variant,
            entity=entity,
            value=str(value),
            ts=ts,
        )

        labels = iter([label])
        self._stub.WriteLabels(labels)

    def close(self):
        """
        Closes the client, closes channel for hosted mode and db for local mode
        """
        self.impl.close()

    @staticmethod
    def _validate_host(host):
        if host.startswith("http://") or host.startswith("https://"):
            raise ValueError("Invalid Host: Host should not contain http or https.")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()



from abc import ABC, abstractmethod
from typing import List, Any, Union

import numpy as np

from ..enums import ScalarType


class ValueType(ABC):
    @abstractmethod
    def scalar(self):
        pass

    @abstractmethod
    def is_vector(self):
        pass


class Scalar(ValueType):
    def __init__(self, scalar_type: ScalarType):
        self.scalar_type = scalar_type

    def scalar(self):
        return self.scalar_type

    def is_vector(self):
        return False


class VectorType(ValueType):
    def __init__(self, scalar_type: ScalarType, dimension: int, is_embedding: bool):
        self.scalar_type = scalar_type
        self.dimension = dimension
        self.is_embedding = is_embedding

    def scalar(self):
        return self.scalar_type

    def is_vector(self):
        return True


class OnlineStore(ABC):
    @abstractmethod
    def get_table(self, feature: str, variant: str):
        pass

    @abstractmethod
    def create_table(self, feature: str, variant: str, entity_type: ValueType):
        pass

    @abstractmethod
    def delete_table(self, feature: str, variant: str):
        pass

    @abstractmethod
    def table_exists(self, name, variant):
        pass

    @abstractmethod
    def close(self):
        pass


class OnlineStoreTable(ABC):
    @abstractmethod
    def set(self, entity: str, value: object):
        pass

    @abstractmethod
    def get(self, entity: str):
        pass


class VectorStoreTable(OnlineStoreTable, ABC):
    @abstractmethod
    def nearest(
        self, feature: str, variant: str, vector: Union[List[float], np.ndarray], k: int
    ):
        pass


class VectorStore(OnlineStore, ABC):
    @abstractmethod
    def create_index(self, feature: str, variant: str, vector_type: VectorType):
        pass

    @abstractmethod
    def delete_index(self, feature: str, variant: str):
        pass



import pytest
from featureform import FilePrefix


@pytest.mark.parametrize(
    "store_type, file_path",
    [
        ("AZURE", "abfss://test_bucket/test_file"),
        ("AZURE", "test_bucket/test_file"),
        ("S3", "s3://test_bucket/test_file"),
        ("S3", "test_bucket/test_file"),
        ("HDFS", "hdfs://test_bucket/test_file"),
        ("HDFS", "test_bucket/test_file"),
        ("GCS", "gs://test_bucket/test_file"),
        ("GCS", "test_bucket/test_file"),
    ],
)
def test_validate_file_scheme(store_type, file_path):
    try:
        FilePrefix.validate(store_type, file_path)
    except Exception as e:
        assert (
            str(e)
            == f"File path '{file_path}' must be a full path. Must start with '{FilePrefix[store_type].prefixes}'"
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: featureform/proto/metadata.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n featureform/proto/metadata.proto\x12\"featureform.serving.metadata.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1egoogle/protobuf/duration.proto\"\x14\n\x04Name\x12\x0c\n\x04name\x18\x01 \x01(\t\"\xbc\x01\n\x0eResourceStatus\x12I\n\x06status\x18\x01 \x01(\x0e\x32\x39.featureform.serving.metadata.proto.ResourceStatus.Status\x12\x15\n\rerror_message\x18\x02 \x01(\t\"H\n\x06Status\x12\r\n\tNO_STATUS\x10\x00\x12\x0b\n\x07\x43REATED\x10\x01\x12\x0b\n\x07PENDING\x10\x02\x12\t\n\x05READY\x10\x03\x12\n\n\x06\x46\x41ILED\x10\x04\"\x98\x01\n\nResourceID\x12\x41\n\x08resource\x18\x01 \x01(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12G\n\rresource_type\x18\x02 \x01(\x0e\x32\x30.featureform.serving.metadata.proto.ResourceType\"\x9b\x01\n\x10SetStatusRequest\x12\x43\n\x0bresource_id\x18\x01 \x01(\x0b\x32..featureform.serving.metadata.proto.ResourceID\x12\x42\n\x06status\x18\x02 \x01(\x0b\x32\x32.featureform.serving.metadata.proto.ResourceStatus\"n\n\x15ScheduleChangeRequest\x12\x43\n\x0bresource_id\x18\x01 \x01(\x0b\x32..featureform.serving.metadata.proto.ResourceID\x12\x10\n\x08schedule\x18\x02 \x01(\t\",\n\x0bNameVariant\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07variant\x18\x02 \x01(\t\"\x07\n\x05\x45mpty\"\x86\x01\n\x07\x46\x65\x61ture\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x42\n\x06status\x18\x02 \x01(\x0b\x32\x32.featureform.serving.metadata.proto.ResourceStatus\x12\x17\n\x0f\x64\x65\x66\x61ult_variant\x18\x03 \x01(\t\x12\x10\n\x08variants\x18\x04 \x03(\t\"4\n\x07\x43olumns\x12\x0e\n\x06\x65ntity\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\x12\n\n\x02ts\x18\x03 \x01(\t\"\x1f\n\x0ePythonFunction\x12\r\n\x05query\x18\x01 \x01(\x0c\"\"\n\x06Stream\x12\x18\n\x10offline_provider\x18\x01 \x01(\t\"\xfa\x06\n\x0e\x46\x65\x61tureVariant\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07variant\x18\x02 \x01(\t\x12?\n\x06source\x18\x03 \x01(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12\x0c\n\x04type\x18\x04 \x01(\t\x12\x0e\n\x06\x65ntity\x18\x05 \x01(\t\x12+\n\x07\x63reated\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\r\n\x05owner\x18\x07 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x08 \x01(\t\x12\x10\n\x08provider\x18\t \x01(\t\x12\x42\n\x06status\x18\n \x01(\x0b\x32\x32.featureform.serving.metadata.proto.ResourceStatus\x12\x45\n\x0ctrainingsets\x18\x0b \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12>\n\x07\x63olumns\x18\x0c \x01(\x0b\x32+.featureform.serving.metadata.proto.ColumnsH\x00\x12\x46\n\x08\x66unction\x18\x11 \x01(\x0b\x32\x32.featureform.serving.metadata.proto.PythonFunctionH\x00\x12<\n\x06stream\x18\x15 \x01(\x0b\x32*.featureform.serving.metadata.proto.StreamH\x00\x12\x30\n\x0clast_updated\x18\r \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x10\n\x08schedule\x18\x0e \x01(\t\x12\x36\n\x04tags\x18\x0f \x01(\x0b\x32(.featureform.serving.metadata.proto.Tags\x12\x42\n\nproperties\x18\x10 \x01(\x0b\x32..featureform.serving.metadata.proto.Properties\x12\x41\n\x04mode\x18\x12 \x01(\x0e\x32\x33.featureform.serving.metadata.proto.ComputationMode\x12\x14\n\x0cis_embedding\x18\x13 \x01(\x08\x12\x11\n\tdimension\x18\x14 \x01(\x05\x42\n\n\x08location\"d\n\nFeatureLag\x12\x0f\n\x07\x66\x65\x61ture\x18\x01 \x01(\t\x12\x0f\n\x07variant\x18\x02 \x01(\t\x12\x0c\n\x04name\x18\x03 \x01(\t\x12&\n\x03lag\x18\x04 \x01(\x0b\x32\x19.google.protobuf.Duration\"\x84\x01\n\x05Label\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x42\n\x06status\x18\x02 \x01(\x0b\x32\x32.featureform.serving.metadata.proto.ResourceStatus\x12\x17\n\x0f\x64\x65\x66\x61ult_variant\x18\x03 \x01(\t\x12\x10\n\x08variants\x18\x04 \x03(\t\"\x80\x05\n\x0cLabelVariant\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07variant\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x0c\n\x04type\x18\x04 \x01(\t\x12?\n\x06source\x18\x05 \x01(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12\x0e\n\x06\x65ntity\x18\x06 \x01(\t\x12+\n\x07\x63reated\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\r\n\x05owner\x18\x08 \x01(\t\x12\x10\n\x08provider\x18\t \x01(\t\x12\x42\n\x06status\x18\n \x01(\x0b\x32\x32.featureform.serving.metadata.proto.ResourceStatus\x12\x45\n\x0ctrainingsets\x18\x0b \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12>\n\x07\x63olumns\x18\x0c \x01(\x0b\x32+.featureform.serving.metadata.proto.ColumnsH\x00\x12<\n\x06stream\x18\x10 \x01(\x0b\x32*.featureform.serving.metadata.proto.StreamH\x00\x12\x36\n\x04tags\x18\r \x01(\x0b\x32(.featureform.serving.metadata.proto.Tags\x12\x42\n\nproperties\x18\x0e \x01(\x0b\x32..featureform.serving.metadata.proto.PropertiesB\n\n\x08location\"\xc3\x04\n\x08Provider\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x0c\n\x04type\x18\x03 \x01(\t\x12\x10\n\x08software\x18\x04 \x01(\t\x12\x0c\n\x04team\x18\x05 \x01(\t\x12\x19\n\x11serialized_config\x18\x06 \x01(\x0c\x12\x42\n\x06status\x18\x07 \x01(\x0b\x32\x32.featureform.serving.metadata.proto.ResourceStatus\x12@\n\x07sources\x18\x08 \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12\x41\n\x08\x66\x65\x61tures\x18\t \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12\x45\n\x0ctrainingsets\x18\n \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12?\n\x06labels\x18\x0b \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12\x36\n\x04tags\x18\x0c \x01(\x0b\x32(.featureform.serving.metadata.proto.Tags\x12\x42\n\nproperties\x18\r \x01(\x0b\x32..featureform.serving.metadata.proto.Properties\"\x8a\x01\n\x0bTrainingSet\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x42\n\x06status\x18\x02 \x01(\x0b\x32\x32.featureform.serving.metadata.proto.ResourceStatus\x12\x17\n\x0f\x64\x65\x66\x61ult_variant\x18\x03 \x01(\t\x12\x10\n\x08variants\x18\x04 \x03(\t\"\xe3\x04\n\x12TrainingSetVariant\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07variant\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\r\n\x05owner\x18\x04 \x01(\t\x12+\n\x07\x63reated\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x10\n\x08provider\x18\x06 \x01(\t\x12\x42\n\x06status\x18\x07 \x01(\x0b\x32\x32.featureform.serving.metadata.proto.ResourceStatus\x12\x41\n\x08\x66\x65\x61tures\x18\x08 \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12>\n\x05label\x18\t \x01(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12\x30\n\x0clast_updated\x18\r \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x10\n\x08schedule\x18\x0e \x01(\t\x12\x44\n\x0c\x66\x65\x61ture_lags\x18\x0f \x03(\x0b\x32..featureform.serving.metadata.proto.FeatureLag\x12\x36\n\x04tags\x18\x10 \x01(\x0b\x32(.featureform.serving.metadata.proto.Tags\x12\x42\n\nproperties\x18\x11 \x01(\x0b\x32..featureform.serving.metadata.proto.Properties\"\xb6\x03\n\x06\x45ntity\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x42\n\x06status\x18\x03 \x01(\x0b\x32\x32.featureform.serving.metadata.proto.ResourceStatus\x12\x41\n\x08\x66\x65\x61tures\x18\x04 \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12?\n\x06labels\x18\x05 \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12\x45\n\x0ctrainingsets\x18\x06 \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12\x36\n\x04tags\x18\x07 \x01(\x0b\x32(.featureform.serving.metadata.proto.Tags\x12\x42\n\nproperties\x18\x08 \x01(\x0b\x32..featureform.serving.metadata.proto.Properties\"\xf1\x02\n\x05Model\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x41\n\x08\x66\x65\x61tures\x18\x03 \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12?\n\x06labels\x18\x04 \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12\x45\n\x0ctrainingsets\x18\x05 \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12\x36\n\x04tags\x18\x06 \x01(\x0b\x32(.featureform.serving.metadata.proto.Tags\x12\x42\n\nproperties\x18\x07 \x01(\x0b\x32..featureform.serving.metadata.proto.Properties\"\xe1\x03\n\x04User\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x42\n\x06status\x18\x02 \x01(\x0b\x32\x32.featureform.serving.metadata.proto.ResourceStatus\x12\x41\n\x08\x66\x65\x61tures\x18\x03 \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12?\n\x06labels\x18\x04 \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12\x45\n\x0ctrainingsets\x18\x05 \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12@\n\x07sources\x18\x06 \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12\x36\n\x04tags\x18\x08 \x01(\x0b\x32(.featureform.serving.metadata.proto.Tags\x12\x42\n\nproperties\x18\t \x01(\x0b\x32..featureform.serving.metadata.proto.Properties\"\x85\x01\n\x06Source\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x42\n\x06status\x18\x02 \x01(\x0b\x32\x32.featureform.serving.metadata.proto.ResourceStatus\x12\x17\n\x0f\x64\x65\x66\x61ult_variant\x18\x03 \x01(\t\x12\x10\n\x08variants\x18\x04 \x03(\t\"\x93\x06\n\rSourceVariant\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07variant\x18\x02 \x01(\t\x12L\n\x0etransformation\x18\x0e \x01(\x0b\x32\x32.featureform.serving.metadata.proto.TransformationH\x00\x12\x46\n\x0bprimaryData\x18\x0f \x01(\x0b\x32/.featureform.serving.metadata.proto.PrimaryDataH\x00\x12\r\n\x05owner\x18\x04 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x12\x10\n\x08provider\x18\x06 \x01(\t\x12+\n\x07\x63reated\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x42\n\x06status\x18\x08 \x01(\x0b\x32\x32.featureform.serving.metadata.proto.ResourceStatus\x12\r\n\x05table\x18\t \x01(\t\x12\x45\n\x0ctrainingsets\x18\n \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12\x41\n\x08\x66\x65\x61tures\x18\x0b \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12?\n\x06labels\x18\x0c \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12\x30\n\x0clast_updated\x18\r \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x10\n\x08schedule\x18\x10 \x01(\t\x12\x36\n\x04tags\x18\x11 \x01(\x0b\x32(.featureform.serving.metadata.proto.Tags\x12\x42\n\nproperties\x18\x12 \x01(\x0b\x32..featureform.serving.metadata.proto.PropertiesB\x0c\n\ndefinition\"\x95\x02\n\x0eTransformation\x12R\n\x11SQLTransformation\x18\x01 \x01(\x0b\x32\x35.featureform.serving.metadata.proto.SQLTransformationH\x00\x12P\n\x10\x44\x46Transformation\x18\x02 \x01(\x0b\x32\x34.featureform.serving.metadata.proto.DFTransformationH\x00\x12M\n\x0fkubernetes_args\x18\x03 \x01(\x0b\x32\x32.featureform.serving.metadata.proto.KubernetesArgsH\x01\x42\x06\n\x04typeB\x06\n\x04\x61rgs\"o\n\x17KubernetesResourceSpecs\x12\x13\n\x0b\x63pu_request\x18\x01 \x01(\t\x12\x11\n\tcpu_limit\x18\x02 \x01(\t\x12\x16\n\x0ememory_request\x18\x03 \x01(\t\x12\x14\n\x0cmemory_limit\x18\x04 \x01(\t\"r\n\x0eKubernetesArgs\x12\x14\n\x0c\x64ocker_image\x18\x01 \x01(\t\x12J\n\x05specs\x18\x02 \x01(\x0b\x32;.featureform.serving.metadata.proto.KubernetesResourceSpecs\"c\n\x11SQLTransformation\x12\r\n\x05query\x18\x01 \x01(\t\x12?\n\x06source\x18\x02 \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\"w\n\x10\x44\x46Transformation\x12\r\n\x05query\x18\x01 \x01(\x0c\x12?\n\x06inputs\x18\x02 \x03(\x0b\x32/.featureform.serving.metadata.proto.NameVariant\x12\x13\n\x0bsource_text\x18\x03 \x01(\t\"_\n\x0bPrimaryData\x12\x44\n\x05table\x18\x01 \x01(\x0b\x32\x33.featureform.serving.metadata.proto.PrimarySQLTableH\x00\x42\n\n\x08location\"\x1f\n\x0fPrimarySQLTable\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x13\n\x04Tags\x12\x0b\n\x03tag\x18\x01 \x03(\t\"+\n\x08Property\x12\x16\n\x0cstring_value\x18\x01 \x01(\tH\x00\x42\x07\n\x05value\"\xbb\x01\n\nProperties\x12N\n\x08property\x18\x01 \x03(\x0b\x32<.featureform.serving.metadata.proto.Properties.PropertyEntry\x1a]\n\rPropertyEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12;\n\x05value\x18\x02 \x01(\x0b\x32,.featureform.serving.metadata.proto.Property:\x02\x38\x01\"\x7f\n\x17StreamingFeatureVariant\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07variant\x18\x02 \x01(\t\x12\x0e\n\x06\x65ntity\x18\x03 \x01(\t\x12\r\n\x05value\x18\x04 \x01(\t\x12&\n\x02ts\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"}\n\x15StreamingLabelVariant\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07variant\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\t\x12\x0e\n\x06\x65ntity\x18\x04 \x01(\t\x12&\n\x02ts\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp*\xc9\x01\n\x0cResourceType\x12\x0b\n\x07\x46\x45\x41TURE\x10\x00\x12\t\n\x05LABEL\x10\x01\x12\x10\n\x0cTRAINING_SET\x10\x02\x12\n\n\x06SOURCE\x10\x03\x12\x13\n\x0f\x46\x45\x41TURE_VARIANT\x10\x04\x12\x11\n\rLABEL_VARIANT\x10\x05\x12\x18\n\x14TRAINING_SET_VARIANT\x10\x06\x12\x12\n\x0eSOURCE_VARIANT\x10\x07\x12\x0c\n\x08PROVIDER\x10\x08\x12\n\n\x06\x45NTITY\x10\t\x12\t\n\x05MODEL\x10\n\x12\x08\n\x04USER\x10\x0b*F\n\x0f\x43omputationMode\x12\x0f\n\x0bPRECOMPUTED\x10\x00\x12\x13\n\x0f\x43LIENT_COMPUTED\x10\x01\x12\r\n\tSTREAMING\x10\x02\x32\x88\x1a\n\x08Metadata\x12h\n\x0cListFeatures\x12).featureform.serving.metadata.proto.Empty\x1a+.featureform.serving.metadata.proto.Feature0\x01\x12u\n\x14\x43reateFeatureVariant\x12\x32.featureform.serving.metadata.proto.FeatureVariant\x1a).featureform.serving.metadata.proto.Empty\x12h\n\x0bGetFeatures\x12(.featureform.serving.metadata.proto.Name\x1a+.featureform.serving.metadata.proto.Feature(\x01\x30\x01\x12}\n\x12GetFeatureVariants\x12/.featureform.serving.metadata.proto.NameVariant\x1a\x32.featureform.serving.metadata.proto.FeatureVariant(\x01\x30\x01\x12\x64\n\nListLabels\x12).featureform.serving.metadata.proto.Empty\x1a).featureform.serving.metadata.proto.Label0\x01\x12q\n\x12\x43reateLabelVariant\x12\x30.featureform.serving.metadata.proto.LabelVariant\x1a).featureform.serving.metadata.proto.Empty\x12\x64\n\tGetLabels\x12(.featureform.serving.metadata.proto.Name\x1a).featureform.serving.metadata.proto.Label(\x01\x30\x01\x12y\n\x10GetLabelVariants\x12/.featureform.serving.metadata.proto.NameVariant\x1a\x30.featureform.serving.metadata.proto.LabelVariant(\x01\x30\x01\x12p\n\x10ListTrainingSets\x12).featureform.serving.metadata.proto.Empty\x1a/.featureform.serving.metadata.proto.TrainingSet0\x01\x12}\n\x18\x43reateTrainingSetVariant\x12\x36.featureform.serving.metadata.proto.TrainingSetVariant\x1a).featureform.serving.metadata.proto.Empty\x12p\n\x0fGetTrainingSets\x12(.featureform.serving.metadata.proto.Name\x1a/.featureform.serving.metadata.proto.TrainingSet(\x01\x30\x01\x12\x85\x01\n\x16GetTrainingSetVariants\x12/.featureform.serving.metadata.proto.NameVariant\x1a\x36.featureform.serving.metadata.proto.TrainingSetVariant(\x01\x30\x01\x12\x66\n\x0bListSources\x12).featureform.serving.metadata.proto.Empty\x1a*.featureform.serving.metadata.proto.Source0\x01\x12s\n\x13\x43reateSourceVariant\x12\x31.featureform.serving.metadata.proto.SourceVariant\x1a).featureform.serving.metadata.proto.Empty\x12\x66\n\nGetSources\x12(.featureform.serving.metadata.proto.Name\x1a*.featureform.serving.metadata.proto.Source(\x01\x30\x01\x12{\n\x11GetSourceVariants\x12/.featureform.serving.metadata.proto.NameVariant\x1a\x31.featureform.serving.metadata.proto.SourceVariant(\x01\x30\x01\x12\x62\n\tListUsers\x12).featureform.serving.metadata.proto.Empty\x1a(.featureform.serving.metadata.proto.User0\x01\x12\x61\n\nCreateUser\x12(.featureform.serving.metadata.proto.User\x1a).featureform.serving.metadata.proto.Empty\x12\x62\n\x08GetUsers\x12(.featureform.serving.metadata.proto.Name\x1a(.featureform.serving.metadata.proto.User(\x01\x30\x01\x12j\n\rListProviders\x12).featureform.serving.metadata.proto.Empty\x1a,.featureform.serving.metadata.proto.Provider0\x01\x12i\n\x0e\x43reateProvider\x12,.featureform.serving.metadata.proto.Provider\x1a).featureform.serving.metadata.proto.Empty\x12j\n\x0cGetProviders\x12(.featureform.serving.metadata.proto.Name\x1a,.featureform.serving.metadata.proto.Provider(\x01\x30\x01\x12g\n\x0cListEntities\x12).featureform.serving.metadata.proto.Empty\x1a*.featureform.serving.metadata.proto.Entity0\x01\x12\x65\n\x0c\x43reateEntity\x12*.featureform.serving.metadata.proto.Entity\x1a).featureform.serving.metadata.proto.Empty\x12g\n\x0bGetEntities\x12(.featureform.serving.metadata.proto.Name\x1a*.featureform.serving.metadata.proto.Entity(\x01\x30\x01\x12\x64\n\nListModels\x12).featureform.serving.metadata.proto.Empty\x1a).featureform.serving.metadata.proto.Model0\x01\x12\x63\n\x0b\x43reateModel\x12).featureform.serving.metadata.proto.Model\x1a).featureform.serving.metadata.proto.Empty\x12\x64\n\tGetModels\x12(.featureform.serving.metadata.proto.Name\x1a).featureform.serving.metadata.proto.Model(\x01\x30\x01\x12t\n\x11SetResourceStatus\x12\x34.featureform.serving.metadata.proto.SetStatusRequest\x1a).featureform.serving.metadata.proto.Empty\x12}\n\x15RequestScheduleChange\x12\x39.featureform.serving.metadata.proto.ScheduleChangeRequest\x1a).featureform.serving.metadata.proto.Empty2\xff\x1a\n\x03\x41pi\x12\x61\n\nCreateUser\x12(.featureform.serving.metadata.proto.User\x1a).featureform.serving.metadata.proto.Empty\x12i\n\x0e\x43reateProvider\x12,.featureform.serving.metadata.proto.Provider\x1a).featureform.serving.metadata.proto.Empty\x12s\n\x13\x43reateSourceVariant\x12\x31.featureform.serving.metadata.proto.SourceVariant\x1a).featureform.serving.metadata.proto.Empty\x12\x65\n\x0c\x43reateEntity\x12*.featureform.serving.metadata.proto.Entity\x1a).featureform.serving.metadata.proto.Empty\x12u\n\x14\x43reateFeatureVariant\x12\x32.featureform.serving.metadata.proto.FeatureVariant\x1a).featureform.serving.metadata.proto.Empty\x12q\n\x12\x43reateLabelVariant\x12\x30.featureform.serving.metadata.proto.LabelVariant\x1a).featureform.serving.metadata.proto.Empty\x12}\n\x18\x43reateTrainingSetVariant\x12\x36.featureform.serving.metadata.proto.TrainingSetVariant\x1a).featureform.serving.metadata.proto.Empty\x12\x63\n\x0b\x43reateModel\x12).featureform.serving.metadata.proto.Model\x1a).featureform.serving.metadata.proto.Empty\x12}\n\x15RequestScheduleChange\x12\x39.featureform.serving.metadata.proto.ScheduleChangeRequest\x1a).featureform.serving.metadata.proto.Empty\x12\x62\n\x08GetUsers\x12(.featureform.serving.metadata.proto.Name\x1a(.featureform.serving.metadata.proto.User(\x01\x30\x01\x12h\n\x0bGetFeatures\x12(.featureform.serving.metadata.proto.Name\x1a+.featureform.serving.metadata.proto.Feature(\x01\x30\x01\x12}\n\x12GetFeatureVariants\x12/.featureform.serving.metadata.proto.NameVariant\x1a\x32.featureform.serving.metadata.proto.FeatureVariant(\x01\x30\x01\x12\x64\n\tGetLabels\x12(.featureform.serving.metadata.proto.Name\x1a).featureform.serving.metadata.proto.Label(\x01\x30\x01\x12y\n\x10GetLabelVariants\x12/.featureform.serving.metadata.proto.NameVariant\x1a\x30.featureform.serving.metadata.proto.LabelVariant(\x01\x30\x01\x12p\n\x0fGetTrainingSets\x12(.featureform.serving.metadata.proto.Name\x1a/.featureform.serving.metadata.proto.TrainingSet(\x01\x30\x01\x12\x85\x01\n\x16GetTrainingSetVariants\x12/.featureform.serving.metadata.proto.NameVariant\x1a\x36.featureform.serving.metadata.proto.TrainingSetVariant(\x01\x30\x01\x12\x66\n\nGetSources\x12(.featureform.serving.metadata.proto.Name\x1a*.featureform.serving.metadata.proto.Source(\x01\x30\x01\x12{\n\x11GetSourceVariants\x12/.featureform.serving.metadata.proto.NameVariant\x1a\x31.featureform.serving.metadata.proto.SourceVariant(\x01\x30\x01\x12j\n\x0cGetProviders\x12(.featureform.serving.metadata.proto.Name\x1a,.featureform.serving.metadata.proto.Provider(\x01\x30\x01\x12g\n\x0bGetEntities\x12(.featureform.serving.metadata.proto.Name\x1a*.featureform.serving.metadata.proto.Entity(\x01\x30\x01\x12\x64\n\tGetModels\x12(.featureform.serving.metadata.proto.Name\x1a).featureform.serving.metadata.proto.Model(\x01\x30\x01\x12h\n\x0cListFeatures\x12).featureform.serving.metadata.proto.Empty\x1a+.featureform.serving.metadata.proto.Feature0\x01\x12\x64\n\nListLabels\x12).featureform.serving.metadata.proto.Empty\x1a).featureform.serving.metadata.proto.Label0\x01\x12p\n\x10ListTrainingSets\x12).featureform.serving.metadata.proto.Empty\x1a/.featureform.serving.metadata.proto.TrainingSet0\x01\x12\x66\n\x0bListSources\x12).featureform.serving.metadata.proto.Empty\x1a*.featureform.serving.metadata.proto.Source0\x01\x12\x62\n\tListUsers\x12).featureform.serving.metadata.proto.Empty\x1a(.featureform.serving.metadata.proto.User0\x01\x12j\n\rListProviders\x12).featureform.serving.metadata.proto.Empty\x1a,.featureform.serving.metadata.proto.Provider0\x01\x12g\n\x0cListEntities\x12).featureform.serving.metadata.proto.Empty\x1a*.featureform.serving.metadata.proto.Entity0\x01\x12\x64\n\nListModels\x12).featureform.serving.metadata.proto.Empty\x1a).featureform.serving.metadata.proto.Model0\x01\x12y\n\rWriteFeatures\x12;.featureform.serving.metadata.proto.StreamingFeatureVariant\x1a).featureform.serving.metadata.proto.Empty(\x01\x12u\n\x0bWriteLabels\x12\x39.featureform.serving.metadata.proto.StreamingLabelVariant\x1a).featureform.serving.metadata.proto.Empty(\x01\x42\'Z%github.com/featureform/metadata/protob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'featureform.proto.metadata_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z%github.com/featureform/metadata/proto'
  _PROPERTIES_PROPERTYENTRY._options = None
  _PROPERTIES_PROPERTYENTRY._serialized_options = b'8\001'
  _globals['_RESOURCETYPE']._serialized_start=7797
  _globals['_RESOURCETYPE']._serialized_end=7998
  _globals['_COMPUTATIONMODE']._serialized_start=8000
  _globals['_COMPUTATIONMODE']._serialized_end=8070
  _globals['_NAME']._serialized_start=137
  _globals['_NAME']._serialized_end=157
  _globals['_RESOURCESTATUS']._serialized_start=160
  _globals['_RESOURCESTATUS']._serialized_end=348
  _globals['_RESOURCESTATUS_STATUS']._serialized_start=276
  _globals['_RESOURCESTATUS_STATUS']._serialized_end=348
  _globals['_RESOURCEID']._serialized_start=351
  _globals['_RESOURCEID']._serialized_end=503
  _globals['_SETSTATUSREQUEST']._serialized_start=506
  _globals['_SETSTATUSREQUEST']._serialized_end=661
  _globals['_SCHEDULECHANGEREQUEST']._serialized_start=663
  _globals['_SCHEDULECHANGEREQUEST']._serialized_end=773
  _globals['_NAMEVARIANT']._serialized_start=775
  _globals['_NAMEVARIANT']._serialized_end=819
  _globals['_EMPTY']._serialized_start=821
  _globals['_EMPTY']._serialized_end=828
  _globals['_FEATURE']._serialized_start=831
  _globals['_FEATURE']._serialized_end=965
  _globals['_COLUMNS']._serialized_start=967
  _globals['_COLUMNS']._serialized_end=1019
  _globals['_PYTHONFUNCTION']._serialized_start=1021
  _globals['_PYTHONFUNCTION']._serialized_end=1052
  _globals['_STREAM']._serialized_start=1054
  _globals['_STREAM']._serialized_end=1088
  _globals['_FEATUREVARIANT']._serialized_start=1091
  _globals['_FEATUREVARIANT']._serialized_end=1981
  _globals['_FEATURELAG']._serialized_start=1983
  _globals['_FEATURELAG']._serialized_end=2083
  _globals['_LABEL']._serialized_start=2086
  _globals['_LABEL']._serialized_end=2218
  _globals['_LABELVARIANT']._serialized_start=2221
  _globals['_LABELVARIANT']._serialized_end=2861
  _globals['_PROVIDER']._serialized_start=2864
  _globals['_PROVIDER']._serialized_end=3443
  _globals['_TRAININGSET']._serialized_start=3446
  _globals['_TRAININGSET']._serialized_end=3584
  _globals['_TRAININGSETVARIANT']._serialized_start=3587
  _globals['_TRAININGSETVARIANT']._serialized_end=4198
  _globals['_ENTITY']._serialized_start=4201
  _globals['_ENTITY']._serialized_end=4639
  _globals['_MODEL']._serialized_start=4642
  _globals['_MODEL']._serialized_end=5011
  _globals['_USER']._serialized_start=5014
  _globals['_USER']._serialized_end=5495
  _globals['_SOURCE']._serialized_start=5498
  _globals['_SOURCE']._serialized_end=5631
  _globals['_SOURCEVARIANT']._serialized_start=5634
  _globals['_SOURCEVARIANT']._serialized_end=6421
  _globals['_TRANSFORMATION']._serialized_start=6424
  _globals['_TRANSFORMATION']._serialized_end=6701
  _globals['_KUBERNETESRESOURCESPECS']._serialized_start=6703
  _globals['_KUBERNETESRESOURCESPECS']._serialized_end=6814
  _globals['_KUBERNETESARGS']._serialized_start=6816
  _globals['_KUBERNETESARGS']._serialized_end=6930
  _globals['_SQLTRANSFORMATION']._serialized_start=6932
  _globals['_SQLTRANSFORMATION']._serialized_end=7031
  _globals['_DFTRANSFORMATION']._serialized_start=7033
  _globals['_DFTRANSFORMATION']._serialized_end=7152
  _globals['_PRIMARYDATA']._serialized_start=7154
  _globals['_PRIMARYDATA']._serialized_end=7249
  _globals['_PRIMARYSQLTABLE']._serialized_start=7251
  _globals['_PRIMARYSQLTABLE']._serialized_end=7282
  _globals['_TAGS']._serialized_start=7284
  _globals['_TAGS']._serialized_end=7303
  _globals['_PROPERTY']._serialized_start=7305
  _globals['_PROPERTY']._serialized_end=7348
  _globals['_PROPERTIES']._serialized_start=7351
  _globals['_PROPERTIES']._serialized_end=7538
  _globals['_PROPERTIES_PROPERTYENTRY']._serialized_start=7445
  _globals['_PROPERTIES_PROPERTYENTRY']._serialized_end=7538
  _globals['_STREAMINGFEATUREVARIANT']._serialized_start=7540
  _globals['_STREAMINGFEATUREVARIANT']._serialized_end=7667
  _globals['_STREAMINGLABELVARIANT']._serialized_start=7669
  _globals['_STREAMINGLABELVARIANT']._serialized_end=7794
  _globals['_METADATA']._serialized_start=8073
  _globals['_METADATA']._serialized_end=11409
  _globals['_API']._serialized_start=11412
  _globals['_API']._serialized_end=14867
# @@protoc_insertion_point(module_scope)



import os.path
import shutil
import stat

import featureform as ff
import pandas as pd
import pytest
from dataclasses import dataclass
from featureform import local, ServingClient

real_path = os.path.realpath(__file__)
dir_path = os.path.dirname(real_path)
SOURCE_FILE = f"{dir_path}/test_files/input_files/transactions.csv"


@dataclass
class SetupFixture:
    transactions_file: str
    serving_client: ServingClient


@pytest.fixture(scope="function")
def setup(tmp_path_factory):
    variant = "cache_test"

    temp_dir = tmp_path_factory.mktemp("test_inputs")
    temp_transactions = temp_dir / "transactions.csv"
    shutil.copy(SOURCE_FILE, temp_transactions)
    transactions = local.register_file(
        name="transactions",
        variant=variant,
        path=str(temp_transactions),
    )

    @local.df_transformation(variant=variant, inputs=[("transactions", variant)])
    def average_user_transaction(transactions):
        """the average transaction amount for a user"""
        return transactions.groupby("CustomerID")["TransactionAmount"].mean()

    user = ff.register_entity("user")

    # Register a column from our transformation as a feature
    average_user_transaction.register_resources(
        entity=user,
        entity_column="CustomerID",
        inference_store=local,
        features=[
            {
                "name": "avg_transactions",
                "variant": variant,
                "column": "TransactionAmount",
                "type": "float32",
            },
        ],
    )

    # Register label from our base Transactions table
    transactions.register_resources(
        entity=user,
        entity_column="CustomerID",
        timestamp_column="Timestamp",
        labels=[
            {
                "name": "fraudulent",
                "variant": variant,
                "column": "IsFraud",
                "type": "bool",
            },
        ],
    )

    ff.register_training_set(
        "fraud_training",
        variant,
        label=("fraudulent", variant),
        features=[("avg_transactions", variant)],
    )

    with ff.Client(local=True) as client:
        client.apply()
        client.training_set("fraud_training", variant)  # this will create the cache

        yield SetupFixture(
            transactions_file=str(temp_transactions), serving_client=client
        )


@pytest.fixture(autouse=True)
def cleanup():
    yield
    shutil.rmtree(".featureform", onerror=del_rw)


def del_rw(action, name, exc):
    if os.path.exists(name):
        os.chmod(name, stat.S_IWRITE)
        os.remove(name)


class TestLocalCache:
    def test_cache_files_are_created(self, setup):
        """
        Sets up local mode with transactions.csv. Registers a transformation, a feature, and a label.
        Ensures all cached files are created
        """

        cache_files = os.listdir(".featureform/cache")

        expected_cache_files = [
            "transformation__average_user_transaction__cache_test.pkl",
            "feature__avg_transactions__cache_test.pkl",
            "label__fraudulent__cache_test.pkl",
            "training_set__fraud_training__cache_test.pkl",
        ]

        assert os.path.exists(".featureform/cache")

        for expected_file in expected_cache_files:
            assert (
                expected_file in cache_files
            ), f"{expected_file} not found in cache_files"

    def test_label_cached_file_is_read_from(self, setup):
        fixture = setup

        # make a change to the cached file
        label_df = pd.read_pickle(
            ".featureform/cache/label__fraudulent__cache_test.pkl"
        )
        label_df["label"] = None
        label_df.to_pickle(".featureform/cache/label__fraudulent__cache_test.pkl")

        label = fixture.serving_client.impl.db.get_label_variant(
            "fraudulent", "cache_test"
        )
        label_df = fixture.serving_client.impl.get_label_dataframe(label)

        # ensure all label values are null
        assert label_df["label"].isnull().all()

    def test_transformation_cached_file_is_read_from(self, setup):
        fixture = setup

        # make a change to the cached file
        transformation_df = pd.read_pickle(
            ".featureform/cache/transformation__average_user_transaction__cache_test.pkl"
        )
        transformation_df["TransactionAmount"] = 0
        transformation_df.to_pickle(
            ".featureform/cache/transformation__average_user_transaction__cache_test.pkl"
        )

        transformation_df = fixture.serving_client.impl.process_transformation(
            "average_user_transaction", "cache_test"
        )

        # ensure all values are 0
        assert transformation_df["TransactionAmount"].all() == 0

    def test_training_set_cached_file_is_read_from(self, setup):
        fixture = setup

        # make a change to the cached file
        training_set_df = pd.read_pickle(
            ".featureform/cache/training_set__fraud_training__cache_test.pkl"
        )
        training_set_df["fraudulent"] = None
        training_set_df.to_pickle(
            ".featureform/cache/training_set__fraud_training__cache_test.pkl"
        )

        training_set = fixture.serving_client.training_set(
            "fraud_training", "cache_test"
        )
        training_set_df = training_set.pandas()

        assert training_set_df["fraudulent"].isnull().all()

    def test_label_cached_files_are_reset(self, setup):
        """
        Ensures all cached files are reset when the files are modified.
        """
        fixture = setup

        # modify the file. This should reset the cache
        transactions = pd.read_csv(fixture.transactions_file)
        transactions["IsFraud"] = None
        transactions.to_csv(fixture.transactions_file, index=False)

        label = fixture.serving_client.impl.db.get_label_variant(
            "fraudulent", "cache_test"
        )
        label_df = fixture.serving_client.impl.get_label_dataframe(label)

        # ensure all values are null
        assert label_df["label"].isnull().all()

    def test_training_set_cached_files_are_reset(self, setup):
        """
        Ensures all cached files are reset when the files are modified.
        """
        fixture = setup

        # modify the file. This should reset the cache
        transactions = pd.read_csv(fixture.transactions_file)
        transactions["IsFraud"] = None
        transactions.to_csv(fixture.transactions_file, index=False)

        training_set = fixture.serving_client.training_set(
            "fraud_training", "cache_test"
        )
        training_set_df = training_set.pandas()

        assert training_set_df["label"].isnull().all()

    def test_transformation_cached_files_are_reset(self, setup):
        """
        Ensures all cached files are reset when the files are modified.
        """
        fixture = setup

        # modify the file. This should reset the cache
        transactions = pd.read_csv(fixture.transactions_file)
        transactions["TransactionAmount"] = 0
        transactions.to_csv(fixture.transactions_file, index=False)

        transformation_df = fixture.serving_client.impl.process_transformation(
            "average_user_transaction", "cache_test"
        )

        # ensure all values are 0
        assert transformation_df.sum() == 0

    def test_feature_cached_files_are_reset(self, setup):
        """
        Ensures all cached files are reset when the files are modified.
        """
        fixture = setup

        # modify the file. This should reset the cache
        transactions = pd.read_csv(fixture.transactions_file)
        transactions["TransactionAmount"] = 0
        transactions.to_csv(fixture.transactions_file, index=False)

        feature = fixture.serving_client.impl.db.get_feature_variant(
            "avg_transactions", "cache_test"
        )
        feature_df = fixture.serving_client.impl.get_feature_dataframe(feature)

        # ensure all values are 0
        assert feature_df["avg_transactions.cache_test"].all() == 0



import sys
import time
from typing import Type, Tuple, List

from dataclasses import dataclass
from featureform.proto.metadata_pb2_grpc import ApiStub

from rich.console import Console
from rich.live import Live
from rich.table import Table
from rich.text import Text

from featureform.resources import (
    Resource,
    Provider,
    FeatureVariant,
    OnDemandFeatureVariant,
    TrainingSetVariant,
    LabelVariant,
    SourceVariant,
    StreamFeature,
    StreamLabel,
)

# maximum number of dots printing when featureform apply for Running...
MAX_NUM_RUNNING_DOTS = 10
SECONDS_BETWEEN_STATUS_CHECKS = 2


def display_statuses(stub: ApiStub, resources: List[Resource], verbose=False):
    StatusDisplayer(stub, resources, verbose=verbose).display()


@dataclass
class DisplayStatus:
    resource_type: Type
    name: str
    variant: str
    status: str = "NO_STATUS"
    error: str = None

    def is_finished(self) -> bool:
        return (
            self.status == "READY"
            or self.status == "FAILED"
            or (
                self.resource_type is Provider and self.status == "NO_STATUS"
            )  # Provider is a special case
        )

    @staticmethod
    def from_resource(resource: Resource):
        variant = getattr(resource, "variant", "")
        return DisplayStatus(
            resource_type=type(resource),
            name=resource.name,
            variant=variant,
            status=resource.status,
            error=resource.error,
        )


class StatusDisplayer:
    did_error: bool = False
    RESOURCE_TYPES_TO_CHECK = {
        FeatureVariant,
        OnDemandFeatureVariant,
        TrainingSetVariant,
        LabelVariant,
        SourceVariant,
        StreamFeature,
        StreamLabel,
        Provider,
    }

    STATUS_TO_COLOR = {
        "READY": "green",
        "CREATED": "green",
        "PENDING": "yellow",
        "NO_STATUS": "white",
        "FAILED": "red",
    }

    def __init__(self, stub: ApiStub, resources: List[Resource], verbose=False):
        self.verbose = verbose
        filtered_resources = filter(
            lambda r: type(r) in self.RESOURCE_TYPES_TO_CHECK, resources
        )
        self.stub = stub

        # A more intuitive way to is to store OrderedDict[Resource, DisplayStatus] but you can't hash Resource easily
        self.resource_to_status_list: List[Tuple[Resource, DisplayStatus]] = []

        for r in filtered_resources:
            self.resource_to_status_list.append((r, DisplayStatus.from_resource(r)))

    def update_display_statuses(self):
        for resource, display_status in self.resource_to_status_list:
            if resource.name == "local-mode":
                continue
            if not display_status.is_finished():
                r = resource.get(self.stub)
                display_status.status = r.status
                display_status.error = r.error
                if r.status == "FAILED":
                    self.did_error = True

    def all_statuses_finished(self) -> bool:
        return all(status.is_finished() for _, status in self.resource_to_status_list)

    def create_error_message(self):
        message = ""
        for _, status in self.resource_to_status_list:
            name = status.name
            message += f"{name}: {status.status} - {status.error}\n"
        return message

    def display(self):
        if not self.resource_to_status_list:
            return

        print()
        console = Console()
        with Live(console=console, auto_refresh=True, screen=False) as live:
            i = 0
            while True:
                self.update_display_statuses()
                finished_running = self.all_statuses_finished()

                dots = "." * (1 + i % MAX_NUM_RUNNING_DOTS)

                title = (
                    f"[green]COMPLETED[/]"
                    if finished_running
                    else f"[yellow]RUNNING{dots}[/]"
                )
                table = Table(
                    title=title,
                    title_justify="left",
                    show_header=True,
                    header_style="bold",
                    box=None,
                )
                table.add_column("Resource Type", width=25)
                table.add_column("Name (Variant)", width=50, no_wrap=True)
                table.add_column("Status", width=10)
                table.add_column("Error", style="red")

                for _, status in self.resource_to_status_list:
                    error = f" {status.error}" if status.error else ""
                    resource_type = status.resource_type.__name__
                    name = status.name
                    status_text = (
                        status.status
                        if status.resource_type is not Provider
                        else "CREATED"
                    )

                    table.add_row(
                        Text(resource_type),
                        Text(f"{name} ({status.variant})"),
                        Text(status_text, style=self.STATUS_TO_COLOR[status_text]),
                        Text(error, style="red"),
                    )

                live.update(table)
                live.refresh()

                if finished_running:
                    # This block is used for testing
                    # Tests check for both stderr and an exception
                    # If we don't throw an exception, then tests will pass even when things fail to register
                    # We also print all the error messages because the table does not get saved when
                    # capturing stdout/stderr
                    if self.verbose and self.did_error:
                        statuses = self.create_error_message()
                        sys.tracebacklimit = 0
                        raise Exception("Some resources failed to create\n" + statuses)
                    break

                i += 1
                time.sleep(SECONDS_BETWEEN_STATUS_CHECKS)



from .sqlite_metadata import SQLiteMetadata
from .resources import Provider, EmptyConfig, PineconeConfig, WeaviateConfig


def get_provider_config(provider_type):
    ptype = {
        "LOCAL_ONLINE": EmptyConfig,
        "PINECONE_ONLINE": PineconeConfig,
        "WEAVIATE_ONLINE": WeaviateConfig,
        "": EmptyConfig,
    }
    return ptype[provider_type]()


def get_provider(name):
    sql = SQLiteMetadata()
    row = sql.get_provider(name, False)
    config = get_provider_config(row["provider_type"]).deserialize(
        row["serialized_config"]
    )

    provider = Provider(
        name=row["name"],
        description=row["description"],
        team=row["team"],
        config=config,
        function=row["provider_type"],
        status=row["status"],
    )
    sql.close()
    return provider



import os
from datetime import timedelta

import pytest
import pandas as pd
import featureform as ff
from featureform import local
from featureform.serving import LocalClientImpl

real_path = os.path.realpath(__file__)
dir_path = os.path.dirname(real_path)
SOURCE_FILE = f"{dir_path}/test_files/input_files/lag_features_testing.csv"


def setup():
    ff.register_user("featureformer").make_default_owner()

    local = ff.register_local()

    test_source_file = local.register_file(
        name="source_file",
        variant="testing",
        description="source file",
        path=SOURCE_FILE,
    )

    @local.df_transformation(
        name=f"source_entity", variant="testing", inputs=[(f"source_file", "testing")]
    )
    def source_entity_transformation(df):
        """the source dataset with entity"""
        df["entity_1"] = "source_entity"
        return df

    @local.df_transformation(
        name=f"source_transformation",
        variant="testing",
        inputs=[(f"source_entity", "testing")],
    )
    def source_transformation(df):
        return df

    entity = ff.register_entity("entity")

    # Register a column from our transformation as a feature
    source_transformation.register_resources(
        entity=entity,
        entity_column="entity_1",
        timestamp_column="timestamp",
        inference_store=local,
        features=[
            {
                "name": "testing_feature",
                "variant": "testing",
                "column": "value",
                "type": "float32",
            },
        ],
    )

    # Register label from our base Transactions table
    source_entity_transformation.register_resources(
        entity=entity,
        entity_column="entity_1",
        timestamp_column="timestamp",
        labels=[
            {
                "name": f"testing_label",
                "variant": "testing",
                "column": "score",
                "type": "float32",
            },
        ],
    )

    ff.register_training_set(
        "testing_training",
        "no_lag_features",
        label=(f"testing_label", "testing"),
        features=[
            (f"testing_feature", "testing"),
        ],
    )

    ff.register_training_set(
        "testing_training",
        "one_lag_features",
        label=(f"testing_label", "testing"),
        features=[
            (f"testing_feature", "testing"),
            {
                "feature": f"testing_feature",
                "variant": "testing",
                "name": "testing_feature_lag_1h",
                "lag": timedelta(hours=1),
            },
        ],
    )

    ff.register_training_set(
        "testing_training",
        "two_lag_features",
        label=(f"testing_label", "testing"),
        features=[
            (f"testing_feature", "testing"),
            {
                "feature": f"testing_feature",
                "variant": "testing",
                "name": "testing_feature_lag_1h",
                "lag": timedelta(hours=1),
            },
            {
                "feature": f"testing_feature",
                "variant": "testing",
                "name": "testing_feature_lag_2h",
                "lag": timedelta(hours=2),
            },
        ],
    )

    ff.register_training_set(
        "testing_training",
        "three_lag_features",
        label=(f"testing_label", "testing"),
        features=[
            (f"testing_feature", "testing"),
            {
                "feature": f"testing_feature",
                "variant": "testing",
                "name": "testing_feature_lag_1h",
                "lag": timedelta(hours=1),
            },
            {
                "feature": f"testing_feature",
                "variant": "testing",
                "name": "testing_feature_lag_2h",
                "lag": timedelta(hours=2),
            },
            {
                "feature": f"testing_feature",
                "variant": "testing",
                "lag": timedelta(seconds=10800),
            },
        ],
    )

    resource_client = ff.ResourceClient(local=True)
    resource_client.apply()


@pytest.fixture()
def source_df():
    df = pd.read_csv(SOURCE_FILE)
    df["timestamp"] = pd.to_datetime(df.timestamp)
    df.rename(
        columns={
            "value": "testing_feature.testing",
            "timestamp": "label_timestamp",
            "score": "label",
        },
        inplace=True,
    )
    return df


@pytest.fixture()
def df_no_lag(source_df):
    return source_df[["testing_feature.testing", "label_timestamp", "label"]]


@pytest.fixture()
def df_one_lag(source_df):
    source_df.set_index("label_timestamp", inplace=True)
    source_df["testing_feature_lag_1h"] = source_df["testing_feature.testing"].shift(
        freq=timedelta(hours=1)
    )
    source_df.reset_index(inplace=True)
    return source_df[
        [
            "testing_feature.testing",
            "testing_feature_lag_1h",
            "label_timestamp",
            "label",
        ]
    ]


@pytest.fixture()
def df_two_lags(source_df):
    source_df.set_index("label_timestamp", inplace=True)
    source_df["testing_feature_lag_1h"] = source_df["testing_feature.testing"].shift(
        freq=timedelta(hours=1)
    )
    source_df["testing_feature_lag_2h"] = source_df["testing_feature.testing"].shift(
        freq=timedelta(hours=2)
    )
    source_df.reset_index(inplace=True)
    return source_df[
        [
            "testing_feature.testing",
            "testing_feature_lag_1h",
            "testing_feature_lag_2h",
            "label_timestamp",
            "label",
        ]
    ]


# @pytest.fixture()
# def df_three_lags(source_df):
#     source_df.set_index("label_timestamp", inplace=True)
#     source_df["testing_feature_lag_1h"] = source_df["testing_feature.testing"].shift(freq=timedelta(hours=1))
#     source_df["testing_feature_lag_2h"] = source_df["testing_feature.testing"].shift(freq=timedelta(hours=2))
#     source_df["testing_feature_testing_lag_3_00_00"] = source_df["testing_feature.testing"].shift(freq=timedelta(hours=3))
#     source_df.reset_index(inplace=True)
#     return source_df[['testing_feature.testing', 'testing_feature_lag_1h', 'testing_feature_lag_2h', 'testing_feature_testing_lag_3_00_00', 'label_timestamp', 'label']]


@pytest.mark.parametrize(
    "training_set_variant, expected_df_name",
    [
        ("no_lag_features", "df_no_lag"),
        ("one_lag_features", "df_one_lag"),
        ("two_lag_features", "df_two_lags"),
        # ("three_lag_features", "df_three_lags"),
    ],
)
def test_local_lag_features(training_set_variant, expected_df_name, request):
    expected_df = request.getfixturevalue(expected_df_name)
    serving_client = ff.ServingClient(local=True)
    dataset = serving_client.training_set(
        "testing_training", training_set_variant, include_label_timestamp=True
    )
    df = dataset.pandas()
    df["label_timestamp"] = pd.to_datetime(df.label_timestamp)

    assert df.equals(
        expected_df
    ), f"The dataframes do not match. Expected: {expected_df.head()}, Got: {df.head()}, Expected Info: {expected_df.info()} Got: {df.info()}"


@pytest.fixture()
def no_lag_feature_sql():
    return "SELECT * FROM source_0"


@pytest.fixture()
def one_lag_feature_sql():
    return """SELECT entity_1, timestamp, "testing_feature.testing", "lag_1h", label FROM (SELECT * FROM (SELECT *, row_number FROM ( SELECT entity_1, timestamp, label, "testing_feature.testing", "lag_1h", ROW_NUMBER() over (PARTITION BY entity_1, label, timestamp ORDER BY timestamp DESC, t0_ts DESC) as row_number FROM (( SELECT * FROM source_0 ) 
                         LEFT OUTER JOIN ( SELECT * FROM ( SELECT entity_1 AS t0_entity, "testing_feature.testing" AS "lag_1h", timestamp AS t0_ts
                         FROM source_0) 
                         ORDER BY t0_ts ASC) t0
                         ON (t0_entity = entity_1 AND DATETIME(t0_ts, "+3600.0 seconds") <= timestamp)
                         ) tt ) WHERE row_number=1 ORDER BY timestamp ASC ))"""


@pytest.fixture()
def two_lag_feature_sql():
    return """SELECT entity_1, timestamp, "testing_feature.testing", "lag_1h", "lag_2h", label FROM (SELECT * FROM (SELECT *, row_number FROM ( SELECT entity_1, timestamp, label, "testing_feature.testing", "lag_1h", "lag_2h", ROW_NUMBER() over (PARTITION BY entity_1, label, timestamp ORDER BY timestamp DESC, t0_ts DESC, t1_ts DESC) as row_number FROM (( SELECT * FROM source_0 ) 
                         LEFT OUTER JOIN ( SELECT * FROM ( SELECT entity_1 AS t0_entity, "testing_feature.testing" AS "lag_1h", timestamp AS t0_ts
                         FROM source_0) 
                         ORDER BY t0_ts ASC) t0
                         ON (t0_entity = entity_1 AND DATETIME(t0_ts, "+3600.0 seconds") <= timestamp)
                         
                         LEFT OUTER JOIN ( SELECT * FROM ( SELECT entity_1 AS t1_entity, "testing_feature.testing" AS "lag_2h", timestamp AS t1_ts
                         FROM source_0) 
                         ORDER BY t1_ts ASC) t1
                         ON (t1_entity = entity_1 AND DATETIME(t1_ts, "+7200.0 seconds") <= timestamp)
                         ) tt ) WHERE row_number=1 ORDER BY timestamp ASC ))"""


@pytest.mark.parametrize(
    "lag_features, feature_columns, entity, label, ts, expected_sql_query",
    [
        ([], "", "", "", "", "no_lag_feature_sql"),
        (
            [
                {
                    "feature_name": "testing_feature",
                    "feature_variant": "testing",
                    "feature_lag": timedelta(hours=1).total_seconds(),
                    "feature_new_name": "lag_1h",
                }
            ],
            ["testing_feature.testing"],
            "entity_1",
            "label",
            "timestamp",
            "one_lag_feature_sql",
        ),
        (
            [
                {
                    "feature_name": "testing_feature",
                    "feature_variant": "testing",
                    "feature_lag": timedelta(hours=1).total_seconds(),
                    "feature_new_name": "lag_1h",
                },
                {
                    "feature_name": "testing_feature",
                    "feature_variant": "testing",
                    "feature_lag": timedelta(hours=2).total_seconds(),
                    "feature_new_name": "lag_2h",
                },
            ],
            ["testing_feature.testing"],
            "entity_1",
            "label",
            "timestamp",
            "two_lag_feature_sql",
        ),
    ],
)
def test_get_lag_features_sql_query(
    lag_features, feature_columns, entity, label, ts, expected_sql_query, request
):
    expected_sql_query = request.getfixturevalue(expected_sql_query)

    local_client = LocalClientImpl()

    lag_sql_query = local_client.get_lag_features_sql_query(
        lag_features, feature_columns, entity, label, ts
    )

    assert lag_sql_query == expected_sql_query



import json
from abc import ABC, abstractmethod
from typing import List

from typeguard import typechecked

from . import SQLiteMetadata
from .resources import (
    Feature,
    LocalConfig,
    Resource,
    FeatureVariant,
    Source,
    SourceVariant,
    Provider,
    TrainingSet,
    Label,
    Model,
    Entity,
    LabelVariant,
    TrainingSetVariant,
    User,
    ResourceColumnMapping,
    TrainingSetFeatures,
)


@typechecked
class MetadataRepository(ABC):
    @abstractmethod
    def create_resource(self, resource: Resource):
        raise NotImplementedError

    @abstractmethod
    def update_resource(self, resource):
        raise NotImplementedError

    @abstractmethod
    def get_feature_variant(self, name: str, variant: str) -> FeatureVariant:
        raise NotImplementedError

    @abstractmethod
    def get_features(self) -> List[Feature]:
        raise NotImplementedError

    @abstractmethod
    def get_label_variant(self, name: str, variant: str) -> LabelVariant:
        raise NotImplementedError

    @abstractmethod
    def get_labels(self) -> List[Label]:
        raise NotImplementedError

    @abstractmethod
    def get_source_variant(self, name: str, variant: str) -> SourceVariant:
        raise NotImplementedError

    @abstractmethod
    def get_sources(self) -> List[Source]:
        raise NotImplementedError

    @abstractmethod
    def get_training_set_variant(self, name: str, variant: str) -> TrainingSetVariant:
        raise NotImplementedError

    @abstractmethod
    def get_training_sets(self) -> List[TrainingSet]:
        raise NotImplementedError

    @abstractmethod
    def get_entity(self, name: str) -> Entity:
        raise NotImplementedError

    @abstractmethod
    def get_entities(self) -> List[Entity]:
        raise NotImplementedError

    @abstractmethod
    def get_model(self, name: str) -> Model:
        raise NotImplementedError

    @abstractmethod
    def get_models(self) -> List[Model]:
        raise NotImplementedError

    @abstractmethod
    def get_provider(self, name: str) -> Provider:
        raise NotImplementedError

    @abstractmethod
    def get_providers(self) -> List[Provider]:
        raise NotImplementedError

    @abstractmethod
    def get_user(self, name: str) -> User:
        raise NotImplementedError

    @abstractmethod
    def get_users(self) -> List[User]:
        raise NotImplementedError

    @abstractmethod
    def get_tags_for_resource(
        self, name: str, variant: str, resource_type: str
    ) -> List[str]:
        raise NotImplementedError

    @abstractmethod
    def get_training_set_features(
        self, name: str, variant: str
    ) -> List[TrainingSetFeatures]:
        raise NotImplementedError

    @abstractmethod
    def get_feature_variants_from_source(
        self, source_name: str, source_variant: str
    ) -> List[FeatureVariant]:
        raise NotImplementedError

    @abstractmethod
    def get_training_set_from_features(
        self, feature_name: str, feature_variant: str
    ) -> List[TrainingSetFeatures]:
        raise NotImplementedError

    @abstractmethod
    def get_label_variants_from_source(
        self, source_name: str, source_variant: str
    ) -> List[LabelVariant]:
        raise NotImplementedError

    @abstractmethod
    def get_training_set_variant_from_label(
        self, label_name: str, label_variant: str
    ) -> List[TrainingSetVariant]:
        raise NotImplementedError


class MetadataRepositoryLocalImpl(MetadataRepository):
    def __init__(self, db: SQLiteMetadata):
        self.db = db

    def create_resource(self, resource: Resource):
        # TODO actual db creation should happen here -- I shouldn't call back to the resource
        resource._create_local(self.db)

    def update_resource(self, resource):
        raise NotImplementedError

    def get_model(self, name: str) -> Model:
        model_row = self.db.get_model(name, should_fetch_tags_properties=True)
        return Model(
            name=model_row["name"],
            description="",
            tags=json.loads(model_row["tags"]) if model_row["tags"] else [],
            properties=json.loads(model_row["properties"])
            if model_row["properties"]
            else {},
        )

    def get_models(self) -> List[Model]:
        model_rows = self.db.query_resource("models", should_fetch_tags_properties=True)
        return [
            Model(
                name=row["name"],
                tags=json.loads(row["tags"]) if row["tags"] else [],
                properties=json.loads(row["properties"]) if row["properties"] else {},
            )
            for row in model_rows
        ]

    def get_feature_variant(self, name, variant) -> FeatureVariant:
        result = self.db.get_feature_variant(name, variant)
        return FeatureVariant(
            created=result["created"],
            name=result["name"],
            variant=result["variant"],
            source=(result["source_name"], result["source_variant"]),
            value_type=result["data_type"],
            is_embedding=bool(result["is_embedding"]),
            dims=result["dimension"],
            entity=result["entity"],
            owner=result["owner"],
            provider=result["provider"],
            status=result["status"],
            location=ResourceColumnMapping(
                result["source_entity"],
                result["source_value"],
                result["source_timestamp"],
            ),
            description=result["description"],
            tags=json.loads(result["tags"]) if result["tags"] else [],
            properties=json.loads(result["properties"]) if result["properties"] else {},
        )

    def get_features(self) -> List[Feature]:
        feature_rows = self.db.query_resource("features")
        features = []
        for feature_row in feature_rows:
            variants = [
                r["variant"]
                for r in self.db.query_resource(
                    "feature_variant", "name", feature_row["name"]
                )
            ]
            features.append(
                Feature(
                    name=feature_row["name"],
                    default_variant=feature_row["default_variant"],
                    variants=variants,
                )
            )

        return features

    def get_label_variant(self, name: str, variant: str) -> LabelVariant:
        result = self.db.get_label_variant(name, variant)
        return LabelVariant(
            name=result["name"],
            variant=result["variant"],
            source=(result["source_name"], result["source_variant"]),
            value_type=result["data_type"],
            created=result["created"],
            entity=result["entity"],
            status=result["status"],
            owner=result["owner"],
            provider=result["provider"],
            location=ResourceColumnMapping(
                result["source_entity"],
                result["source_value"],
                result["source_timestamp"],
            ),
            description=result["description"],
            tags=json.loads(result["tags"]) if result["tags"] else [],
            properties=json.loads(result["properties"]) if result["properties"] else {},
        )

    def get_labels(self) -> List[Label]:
        label_rows = self.db.query_resource("labels")
        labels = []
        for label_row in label_rows:
            variants = [
                r["variant"]
                for r in self.db.query_resource(
                    "label_variant", "name", label_row["name"]
                )
            ]
            labels.append(
                Label(
                    name=label_row["name"],
                    default_variant=label_row["default_variant"],
                    variants=variants,
                )
            )

        return labels

    def get_source_variant(self, name: str, variant: str) -> SourceVariant:
        result = self.db.get_source_variant(name, variant)
        return SourceVariant(
            created=result["created"],
            name=result["name"],
            definition=str(result["definition"]),
            variant=result["variant"],
            owner=result["owner"],
            provider=result["provider"],
            source_type=result["source_type"],
            transformation=result["transformation"],
            description=result["description"],
            tags=json.loads(result["tags"]) if result["tags"] else [],
            properties=json.loads(result["properties"]) if result["properties"] else {},
            status=result["status"],
        )

    def get_sources(self) -> List[Source]:
        source_rows = self.db.query_resource("sources")
        sources = []
        for source_row in source_rows:
            variants = [
                r["variant"]
                for r in self.db.query_resource(
                    "source_variant", "name", source_row["name"]
                )
            ]
            sources.append(
                Source(
                    name=source_row["name"],
                    default_variant=source_row["default_variant"],
                    variants=variants,
                )
            )

        return sources

    def get_source_variant_text(self, name: str, variant: str) -> str:
        result = self.db.get_source_variant_text(name, variant)
        return result

    def get_training_set_variant(self, name: str, variant: str) -> TrainingSetVariant:
        result = self.db.get_training_set_variant(name, variant)
        ts_feature_rows = self.db.get_training_set_features(name, variant)
        feature_name_variants = [
            (r["feature_name"], r["feature_variant"]) for r in ts_feature_rows
        ]
        return TrainingSetVariant(  # does local mode use provider for TS
            created=result["created"],
            name=result["name"],
            variant=result["variant"],
            owner=result["owner"],
            label=(result["label_name"], result["label_variant"]),
            features=feature_name_variants,
            description=result["description"],
            tags=json.loads(result["tags"]) if result["tags"] else [],
            properties=json.loads(result["properties"]) if result["properties"] else {},
            status=result["status"],
        )

    def get_training_set(self, name: str) -> TrainingSet:
        result = self.db.get_training_set(name, should_fetch_tags_properties=False)
        return TrainingSet(
            name=result["name"],
            default_variant=result["default_variant"],
            variants=[
                r["variant"]
                for r in self.db.query_resource(
                    "training_set_variant", "name", result["name"]
                )
            ],
        )

    def get_training_sets(self) -> List[TrainingSet]:
        training_set_rows = self.db.query_resource("training_sets")
        training_sets = []
        for ts_row in training_set_rows:
            variants = [
                r["variant"]
                for r in self.db.query_resource(
                    "training_set_variant", "name", ts_row["name"]
                )
            ]
            training_sets.append(
                TrainingSet(
                    name=ts_row["name"],
                    default_variant=ts_row["default_variant"],
                    variants=variants,
                )
            )

        return training_sets

    def get_entity(self, name: str) -> Entity:
        entity_row = self.db.get_entity(name, should_fetch_tags_properties=True)
        return Entity(
            name=entity_row["name"],
            description=entity_row["description"],
            status=entity_row["status"],
            tags=json.loads(entity_row["tags"]) if entity_row["tags"] else [],
            properties=json.loads(entity_row["properties"])
            if entity_row["properties"]
            else {},
        )

    def get_entities(self) -> List[Entity]:
        entity_rows = self.db.query_resource(
            "entities", should_fetch_tags_properties=True
        )
        return [
            Entity(
                name=row["name"],
                description=row["description"],
                status=row["status"],
                tags=json.loads(row["tags"]) if row["tags"] else [],
                properties=json.loads(row["properties"]) if row["properties"] else {},
            )
            for row in entity_rows
        ]

    def get_provider(self, name: str) -> Provider:
        result = self.db.get_provider(name, should_fetch_tags_properties=True)
        return Provider(
            name=result["name"],
            description=result["description"],
            team=result["team"],
            config=LocalConfig(),
            function="",  # look into this
            status=result["status"],
            tags=json.loads(result["tags"]) if result["tags"] else [],
            properties=json.loads(result["properties"]) if result["properties"] else {},
        )

    def get_providers(self) -> List[Provider]:
        provider_rows = self.db.query_resource(
            "providers", should_fetch_tags_properties=True
        )
        return [
            Provider(
                name=row["name"],
                description=row["description"],
                team=row["team"],
                config=LocalConfig(),
                function="",  # look into this
                status=row["status"],
                tags=json.loads(row["tags"]) if row["tags"] else [],
                properties=json.loads(row["properties"]) if row["properties"] else {},
            )
            for row in provider_rows
        ]

    def get_user(self, name: str) -> User:
        result = self.db.get_user(name, should_fetch_tags_properties=True)
        return User(
            name=result["name"],
            status=result["status"],
            tags=json.loads(result["tags"]) if result["tags"] else [],
            properties=json.loads(result["properties"]) if result["properties"] else {},
        )

    def get_users(self) -> List[User]:
        user_rows = self.db.query_resource("users", should_fetch_tags_properties=True)
        return [
            User(
                name=row["name"],
                status=row["status"],
                tags=json.loads(row["tags"]) if row["tags"] else [],
                properties=json.loads(row["properties"]) if row["properties"] else {},
            )
            for row in user_rows
        ]

    def get_tags_for_resource(self, name, variant, resource_type) -> List[str]:
        row_data = self.db.get_tags(
            name=name, variant=variant, resource_type=resource_type
        )
        if len(row_data):
            return row_data[0][0]
        else:
            return []

    def get_training_set_features(
        self, name: str, variant: str
    ) -> List[TrainingSetFeatures]:
        db_result = self.db.get_training_set_features(name=name, variant=variant)
        tsf_list = []
        for row in db_result:
            tsf_list.append(
                TrainingSetFeatures(
                    training_set_name=row["training_set_name"],
                    training_set_variant=row["training_set_variant"],
                    feature_name=row["feature_name"],
                    feature_variant=row["feature_variant"],
                )
            )

        return tsf_list

    def get_feature_variants_from_source(
        self, source_name: str, source_variant: str
    ) -> List[FeatureVariant]:
        db_result = self.db.get_feature_variants_from_source(
            name=source_name, variant=source_variant
        )
        feature_variant_list = []
        for row in db_result:
            feature_variant_list.append(
                FeatureVariant(
                    created=row["created"],
                    name=row["name"],
                    variant=row["variant"],
                    source=(row["source_name"], row["source_variant"]),
                    value_type=row["data_type"],
                    is_embedding=bool(row["is_embedding"]),
                    dims=row["dimension"],
                    entity=row["entity"],
                    owner=row["owner"],
                    provider=row["provider"],
                    status=row["status"],
                    location=ResourceColumnMapping(
                        row["source_entity"],
                        row["source_value"],
                        row["source_timestamp"],
                    ),
                    description=row["description"],
                    tags=json.loads(row["tags"]) if row["tags"] else [],
                    properties=json.loads(row["properties"])
                    if row["properties"]
                    else {},
                )
            )

        return feature_variant_list

    def get_training_set_from_features(
        self, feature_name: str, feature_variant: str
    ) -> List[TrainingSetFeatures]:
        db_result = self.db.get_training_set_from_features(
            name=feature_name, variant=feature_variant
        )
        tsf_list = []
        for row in db_result:
            tsf_list.append(
                TrainingSetFeatures(
                    training_set_name=row["training_set_name"],
                    training_set_variant=row["training_set_variant"],
                    feature_name=row["feature_name"],
                    feature_variant=row["feature_variant"],
                )
            )

        return tsf_list

    def get_label_variants_from_source(
        self, source_name: str, source_variant: str
    ) -> List[LabelVariant]:
        db_result = self.db.get_label_variants_from_source(
            name=source_name, variant=source_variant
        )
        label_variant_list = []
        for row in db_result:
            label_variant_list.append(
                LabelVariant(
                    name=row["name"],
                    variant=row["variant"],
                    source=(row["source_name"], row["source_variant"]),
                    value_type=row["data_type"],
                    created=row["created"],
                    entity=row["entity"],
                    status=row["status"],
                    owner=row["owner"],
                    provider=row["provider"],
                    location=ResourceColumnMapping(
                        row["source_entity"],
                        row["source_value"],
                        row["source_timestamp"],
                    ),
                    description=row["description"],
                    tags=json.loads(row["tags"]) if row["tags"] else [],
                    properties=json.loads(row["properties"])
                    if row["properties"]
                    else {},
                )
            )

        return label_variant_list

    def get_training_set_variant_from_label(
        self, label_name: str, label_variant: str
    ) -> List[TrainingSetVariant]:
        db_result = self.db.get_training_set_variant_from_label(
            name=label_name, variant=label_variant
        )
        training_set_variant_list = []
        for row in db_result:
            ts_feature_rows = self.db.get_training_set_features(
                row["name"], row["variant"]
            )
            feature_name_variants = [
                (r["feature_name"], r["feature_variant"]) for r in ts_feature_rows
            ]
            training_set_variant_list.append(
                TrainingSetVariant(
                    created=row["created"],
                    name=row["name"],
                    variant=row["variant"],
                    owner=row["owner"],
                    label=(row["label_name"], row["label_variant"]),
                    features=feature_name_variants,
                    description=row["description"],
                    tags=json.loads(row["tags"]) if row["tags"] else [],
                    properties=json.loads(row["properties"])
                    if row["properties"]
                    else {},
                    status=row["status"],
                )
            )

        return training_set_variant_list



from .enums import ComputationMode


class InvalidTrainingSetFeatureComputationMode(Exception):
    def __init__(
        self,
        feature_name,
        feature_variant,
        mode=ComputationMode.CLIENT_COMPUTED.value,
        message=None,
    ):
        if message is None:
            message = (
                f"Feature '{feature_name}:{feature_variant}' is on demand. "
                f"Cannot use {mode} features for training sets. "
            )

        Exception.__init__(self, message)


class FeatureNotFound(Exception):
    def __init__(self, feature_name, feature_variant, message=None):
        error_message = f"Feature '{feature_name}:{feature_variant}' not found. Verify that the feature is registered."

        if message is not None:
            error_message = f"{error_message} {message}"

        Exception.__init__(self, error_message)


class LabelNotFound(Exception):
    def __init__(self, label_name, label_variant, message=None):
        error_message = f"Label '{label_name}:{label_variant}' not found. Verify that the label is registered."
        if message is not None:
            error_message = f"{error_message} {message}"

        Exception.__init__(self, error_message)


class InvalidSQLQuery(Exception):
    def __init__(self, query, message=None):
        error_message = f"Invalid SQL query. Query: ' {query} '"
        if message is not None:
            error_message = f"{error_message} {message}"

        Exception.__init__(self, error_message)



import pandas as pd
import os
from ..register import ScalarType
from .online_store import OnlineStore, OnlineStoreTable, ValueType


class LocalFileTable(OnlineStoreTable):
    def __init__(self, name, variant, root_path, stype):
        self.name = name
        self.variant = variant
        self.root_path = root_path
        self.dir = f"{root_path}/{name}/"
        self.filepath = f"{root_path}/{name}/{variant}.csv"
        self.stype = stype
        self.update_df = self.__init_update_df()

        if not os.path.exists(self.dir):
            os.makedirs(self.dir, exist_ok=True)

        if not os.path.exists(self.filepath):
            self.__init_table()
        else:
            self.df = pd.read_csv(self.filepath, header=0, names=["entity", "value"])

    def __init_table(self):
        self.df = pd.DataFrame(data={"entity": [], "value": []}).astype(
            {"entity": "string", "value": self.stype}
        )
        self.df.to_csv(self.filepath, index=False)

    def __init_update_df(self):
        df = pd.DataFrame(columns=["entity", "value"]).astype(
            {"entity": "str", "value": self.stype}
        )
        return df

    def set(self, key, value):
        df = pd.DataFrame(data={"entity": [key], "value": [value]})
        try:
            df = df.astype({"entity": "str", "value": self.stype})
        except:
            raise ValueError(f"Could not cast feature to type {self.stype}")
        self.df = self.df.append(df)

    def set_batch(self, df):
        columns = df.columns
        df = (
            df.reset_index(drop=True)
            .rename(columns={columns[0]: "entity", columns[1]: "value"})
            .drop_duplicates("entity", keep="last")
            .sort_values("entity", ascending=False)
        )
        df.to_csv(self.filepath, mode="w", index=False, header=True)

    def get(self, key):
        # The most efficient way to perform the upsert at the moment is to
        # create an array with new values to be inserted then process them
        # all at the same time in memory.
        self.flush()
        df = self.df
        try:
            value = df[df["entity"] == key]["value"].astype(self.stype).values[0]
        except IndexError:
            raise ValueError(
                f"Could not find value for feature: {self.name} ({self.variant}) for: {key}"
            ) from None
        if self.stype == "datetime64[ns]":
            return pd.to_datetime(value)
        return value

    def flush(self):
        self.df = (
            pd.concat([self.df, self.update_df])
            .drop_duplicates("entity", keep="last")
            .sort_values("entity", ascending=False)
            .reset_index(drop=True)
        )
        self.df.to_csv(self.filepath, mode="w", index=False, header=True)


class LocalFileStore(OnlineStore):
    def __init__(self, config):
        self.path = ".featureform/features"
        self.type_table = f"{self.path}/types.csv"
        if not os.path.exists(self.path):
            os.makedirs(self.path, exist_ok=True)

    def create_table(self, name, variant, entity_type: ValueType):
        if os.path.exists(f"{self.path}/{name}/{variant}.csv"):
            raise ValueError(f"Table {name} with variant {variant} already exists")

        self.__set_type(name, variant, entity_type)

        table = LocalFileTable(
            name, variant, self.path, self.__convert_type(entity_type.scalar())
        )
        return table

    def __set_type(self, name, variant, entity_type: ValueType):
        header = True
        if os.path.exists(self.type_table):
            header = False

        types_df = pd.DataFrame(
            data={
                "name": [name],
                "variant": [variant],
                "type": [entity_type.scalar().value],
            }
        ).astype({"name": "string", "variant": "string", "type": "string"})

        types_df.to_csv(self.type_table, mode="a", header=header, index=False)

    def get_table(self, name, variant):
        if not os.path.exists(f"{self.path}/{name}/{variant}.csv"):
            raise ValueError(f"Table {name} with variant {variant} does not exist")

        stype = self.__get_type(name, variant)

        return LocalFileTable(name, variant, self.path, self.__convert_type(stype))

    def __get_type(self, name, variant) -> ScalarType:
        df = pd.read_csv(self.type_table)
        matched_rows = df.loc[
            (df["name"] == str(name)) & (df["variant"] == str(variant))
        ]
        if len(matched_rows) == 0:
            raise ValueError(
                f"Could not get type for table {name} with variant {variant}"
            )

        return ScalarType(matched_rows["type"].values[0])

    def table_exists(self, name, variant):
        return os.path.exists(f"{self.path}/{name}/{variant}.csv")

    def __convert_type(self, stype: ScalarType):
        types = {
            ScalarType.NIL: "",
            ScalarType.INT: "int64",
            ScalarType.INT32: "int64",
            ScalarType.INT64: "int64",
            ScalarType.FLOAT32: "float64",
            ScalarType.FLOAT64: "float64",
            ScalarType.STRING: "object",
            ScalarType.BOOL: "bool",
            ScalarType.DATETIME: "datetime64[ns]",
        }
        return types[stype]

    def close(self):
        pass

    def delete_table(self, feature: str, variant: str):
        pass



from featureform.providers import (
    get_provider,
    OnlineStore,
    OnlineStoreTable,
    LocalFileStore,
)
from featureform.register import ScalarType
import pytest
import uuid
from datetime import datetime


@pytest.mark.parametrize(
    "provider",
    ["file"],
)
class TestOnlineProvider:
    def test_get_provider(self, provider):
        assert isinstance(get_provider(provider), LocalFileStore)
        with pytest.raises(NotImplementedError):
            get_provider("something else")

    @pytest.mark.parametrize(
        "name,variant,t",
        [(uuid.uuid4(), uuid.uuid4(), ScalarType.INT)],
    )
    def test_create_table(self, provider, name, variant, t):
        store = get_provider(provider)
        assert isinstance(store, OnlineStore)
        table = store.create_table(name, variant, t)
        assert isinstance(table, OnlineStoreTable)

    @pytest.mark.parametrize(
        "name,variant,t",
        [(uuid.uuid4(), uuid.uuid4(), ScalarType.INT)],
    )
    def test_get_table(self, provider, name, variant, t):
        store = get_provider(provider)
        store.create_table(name, variant, t)
        assert isinstance(store, OnlineStore)
        table = store.get_table(name, variant)
        assert isinstance(table, OnlineStoreTable)

    @pytest.mark.parametrize(
        "name,variant,values,t",
        [
            (
                uuid.uuid4(),
                uuid.uuid4(),
                [("a", "one"), ("b", "two"), ("c", "three")],
                ScalarType.STRING,
            ),
            (
                uuid.uuid4(),
                uuid.uuid4(),
                [("a", 1), ("b", 2), ("c", 3)],
                ScalarType.INT,
            ),
            (
                uuid.uuid4(),
                uuid.uuid4(),
                [("a", 1.0), ("b", 2.0), ("c", 3.0)],
                ScalarType.FLOAT64,
            ),
            (
                uuid.uuid4(),
                uuid.uuid4(),
                [("a", True), ("b", False), ("c", True)],
                ScalarType.BOOL,
            ),
            (
                uuid.uuid4(),
                uuid.uuid4(),
                [("a", datetime.now()), ("b", datetime.now()), ("c", datetime.now())],
                ScalarType.DATETIME,
            ),
        ],
    )
    def test_set_get_value(self, provider, name, variant, values, t):
        store = get_provider(provider)
        table = store.create_table(name, variant, t)
        for value in values:
            table.set(value[0], value[1])
        for value in values:
            assert table.get(value[0]) == value[1], f"Failed for entity {value[0]}"



from .format import *
from .sqlite_metadata import *

cutoff_length = 25


def search_local(phrase):
    db = SQLiteMetadata()
    results = db.search(phrase=phrase)

    if len(results) == 0:
        print(f"Search phrase {phrase} returned no results.")
        return []
    else:
        return results



from .sqlite_metadata import *
from .format import *
from enum import Enum

cutoff_length = 60


def get_sorted_list(resource_type, variant=False):
    db = SQLiteMetadata()
    local_list_args = {
        "entity": ["entities"],
        "user": ["users"],
        "model": ["models"],
        "provider": ["providers"],
        "feature": ["features", "feature_variant"],
        "label": ["labels", "label_variant"],
        "source": ["sources", "source_variant"],
        "training-set": ["training_sets", "training_set_variant"],
    }
    if variant:
        received = db.get_type_table(local_list_args[resource_type][1])
        res = sorted([r for r in received], key=lambda x: x["name"])
    else:
        received = db.get_type_table(local_list_args[resource_type][0])
        res = sorted([r for r in received], key=lambda x: x["name"])
    return res


class ColumnName(Enum):
    NAME = 0
    STATUS = 1
    VARIANT = 2
    DESCRIPTION = 3


def list_conditions(x, clm_name, default_variant=None):
    if clm_name == "variant" and default_variant == x[clm_name]:
        return f"{x[clm_name]} (default)"
    if clm_name == "description":
        return x[clm_name][:cutoff_length]
    return x[clm_name]


def list_local(resource_type, fields):
    field_to_clm = {
        ColumnName.NAME: "name",
        ColumnName.VARIANT: "variant",
        ColumnName.STATUS: "status",
        ColumnName.DESCRIPTION: "description",
    }
    clm_names = [field_to_clm[field] for field in fields]
    format_rows(*[field.name for field in fields])
    res = get_sorted_list(resource_type)
    if "variant" in clm_names:
        res_variants = get_sorted_list(resource_type, True)
        for r in res:
            for v in filter(lambda x: x["name"] == r["name"], res_variants):
                format_list = [
                    list_conditions(v, clm_name, r["default_variant"])
                    for clm_name in clm_names
                ]
                format_rows(*format_list)
    else:
        for r in res:
            format_list = [list_conditions(r, clm_name) for clm_name in clm_names]
            format_rows(*format_list)
    return res



import os
import shutil
import stat
from dataclasses import replace

from featureform import SQLiteMetadata, FeatureVariant, ScalarType
from featureform.metadata_repository import (
    MetadataRepository,
    MetadataRepositoryLocalImpl,
)

import pytest

from featureform.resources import (
    ResourceLocation,
    Feature,
    User,
    LabelVariant,
    Provider,
    LocalConfig,
    SourceVariant,
    PrimaryData,
    SQLTable,
    Label,
    Model,
    Entity,
    TrainingSetVariant,
    TrainingSet,
)

test_user = User(
    name="test_user", tags=["tag1"], properties={"key1": "value1"}, status="ready"
)

model = Model(name="model")

entity = Entity(name="entity", description="fake description", status="ready")

provider = Provider(
    name="provider",
    description="fake description",
    team="fake_team",
    config=LocalConfig(),
    function="",
    status="ready",
)

resource_location = ResourceLocation(
    entity="Dummy Entity", value="Dummy Value", timestamp="12345"
)
feature_variant = FeatureVariant(
    created=None,
    name="feature",
    source=("source", "variant1"),
    value_type=ScalarType.FLOAT32.value,
    entity="fake_entity",
    owner="fake_owner",
    provider="fake_provider",
    location=resource_location,
    description="fake description",
    variant="variant1",
    status="ready",
    is_embedding=False,
    dims=0,
    tags=["tag1"],
    properties={"key1": "value1"},
)

source_variant = SourceVariant(
    created=None,
    name="source",
    variant="variant1",
    source_type="Source",
    transformation="PRIMARY",
    definition=PrimaryData(location=SQLTable(name="")),
    owner="fake_owner",
    provider="fake_provider",
    description="fake description",
    tags=[],
    properties={},
)

label_variant = LabelVariant(
    created=None,
    name="label",
    source=("source", "variant1"),
    value_type=ScalarType.FLOAT32.value,
    entity="fake_entity",
    owner="fake_owner",
    provider="fake_provider",
    location=resource_location,
    description="fake description",
    variant="variant1",
    status="ready",
    tags=["tag1"],
    properties={"key1": "value1"},
)

training_set_variant = TrainingSetVariant(
    created=None,
    name="training_set",
    owner="fake_owner",
    label=("label", "variant1"),
    features=[("feature", "variant1")],
    description="fake description",
    variant="variant1",
    tags=["tag1"],
    properties={"key1": "value1"},
    status="ready",
)


@pytest.fixture(autouse=True)
def cleanup():
    yield
    shutil.rmtree(".featureform", onerror=del_rw)


def del_rw(action, name, exc):
    if os.path.exists(name):
        os.chmod(name, stat.S_IWRITE)
        os.remove(name)


@pytest.fixture(scope="function")
def setup() -> MetadataRepository:
    with SQLiteMetadata() as db:
        yield MetadataRepositoryLocalImpl(db)


@pytest.mark.local
def test_create_model(setup: MetadataRepository):
    repo = setup
    repo.create_resource(model)
    assert repo.get_model("model") == model


@pytest.mark.local
def test_get_models(setup: MetadataRepository):
    repo = setup
    repo.create_resource(model)
    model2 = replace(model, name="fake_model2")
    repo.create_resource(model2)

    models = repo.get_models()
    assert len(models) == 2  # there's one more for local mode
    assert model in models
    assert model2 in models


@pytest.mark.local
def test_create_entity(setup: MetadataRepository):
    repo = setup
    repo.create_resource(entity)
    assert repo.get_entity("entity") == entity


@pytest.mark.local
def test_get_entities(setup: MetadataRepository):
    repo = setup
    repo.create_resource(entity)
    entity2 = replace(entity, name="entity2")
    repo.create_resource(entity2)

    entities = repo.get_entities()
    assert len(entities) == 2  # there's one more for local mode
    assert entity in entities
    assert entity2 in entities


@pytest.mark.local
def test_create_provider(setup: MetadataRepository):
    repo = setup
    repo.create_resource(provider)
    assert repo.get_provider("provider") == provider


@pytest.mark.local
def test_get_providers(setup: MetadataRepository):
    repo = setup
    repo.create_resource(provider)
    provider2 = replace(provider, name="fake_provider2")
    repo.create_resource(provider2)

    providers = repo.get_providers()
    assert len(providers) == 2  # there's one more for local mode
    assert provider in providers
    assert provider2 in providers


@pytest.mark.local
def test_create_source_variant(setup: MetadataRepository):
    repo = setup
    repo.create_resource(source_variant)
    found_variant = repo.get_source_variant("source", "variant1")
    found_variant.created = None
    assert found_variant == source_variant


@pytest.mark.local
def test_get_sources(setup: MetadataRepository):
    repo = setup
    repo.create_resource(source_variant)
    source2 = replace(source_variant, variant="variant2")
    repo.create_resource(source2)

    sources = repo.get_sources()
    assert len(sources) == 1
    assert sources[0].variants == ["variant1", "variant2"]


@pytest.mark.local
def test_create_label_variant(setup: MetadataRepository):
    repo = setup
    repo.create_resource(label_variant)
    found_variant = repo.get_label_variant("label", "variant1")
    found_variant.created = None
    assert found_variant == label_variant


@pytest.mark.local
def test_get_labels(setup: MetadataRepository):
    repo = setup
    repo.create_resource(label_variant)
    label_variant2 = replace(label_variant, variant="variant2")
    repo.create_resource(label_variant2)

    labels = repo.get_labels()
    assert len(labels) == 1
    assert labels[0] == Label(
        name="label",
        default_variant="variant1",
        variants=["variant1", "variant2"],
    )


@pytest.mark.local
def test_create_feature_variant(setup: MetadataRepository):
    repo = setup
    repo.create_resource(feature_variant)
    found_variant = repo.get_feature_variant("feature", "variant1")
    found_variant.created = None
    assert found_variant == feature_variant


@pytest.mark.local
def test_get_features(setup: MetadataRepository):
    repo = setup
    repo.create_resource(feature_variant)
    feature_variant2 = replace(feature_variant, variant="variant2")
    repo.create_resource(feature_variant2)

    features = repo.get_features()
    assert len(features) == 1
    assert features[0] == Feature(
        name="feature",
        default_variant="variant1",
        variants=["variant1", "variant2"],
    )


@pytest.mark.local
def test_create_training_set_variant(setup: MetadataRepository):
    repo = setup
    repo.create_resource(label_variant)
    repo.create_resource(feature_variant)
    repo.create_resource(training_set_variant)
    found_variant = repo.get_training_set_variant("training_set", "variant1")
    found_variant.created = None
    assert found_variant == training_set_variant


@pytest.mark.local
def test_get_training_sets(setup: MetadataRepository):
    repo = setup
    repo.create_resource(label_variant)
    repo.create_resource(feature_variant)
    repo.create_resource(training_set_variant)
    training_set_variant2 = replace(training_set_variant, variant="variant2")
    repo.create_resource(training_set_variant2)

    training_sets = repo.get_training_sets()
    assert len(training_sets) == 1
    assert training_sets[0] == TrainingSet(
        name="training_set",
        default_variant="variant1",
        variants=["variant1", "variant2"],
    )


@pytest.mark.local
def test_create_user(setup: MetadataRepository):
    repo: MetadataRepository = setup
    repo.create_resource(test_user)
    retrieved_user = repo.get_user(test_user.name)
    assert retrieved_user == test_user


@pytest.mark.local
def test_get_users(setup: MetadataRepository):
    repo: MetadataRepository = setup

    user1 = test_user
    user2 = replace(test_user, name="test_user2")
    user3 = replace(test_user, name="test_user3")

    users = [user1, user2, user3]

    for user in users:
        repo.create_resource(user)

    retrieved_users = repo.get_users()
    assert len(retrieved_users) == 3
    assert retrieved_users == users


@pytest.mark.local
def test_get_feature_variants_from_source(setup: MetadataRepository):
    repo = setup
    f_variant = replace(feature_variant, source=("custom_name", "custom_variant"))
    repo.create_resource(f_variant)
    found_list = repo.get_feature_variants_from_source(
        source_name=f_variant.source[0], source_variant=f_variant.source[1]
    )
    assert len(found_list) == 1


@pytest.mark.local
def test_get_label_variants_from_source(setup: MetadataRepository):
    repo = setup
    l_variant = replace(label_variant, source=("custom_name", "custom_variant"))
    repo.create_resource(l_variant)
    found_list = repo.get_label_variants_from_source(
        source_name=l_variant.source[0], source_variant=l_variant.source[1]
    )
    assert len(found_list) == 1



import pytest

from featureform.resources import DatabricksCredentials, EMRCredentials


@pytest.mark.parametrize(
    "username,password,host,token,cluster_id",
    [
        ("a", "b", "", "", "c"),
        ("", "", "a", "b", "c"),
        pytest.param("a", "b", "a", "b", "c", marks=pytest.mark.xfail),
        pytest.param("", "", "", "", "c", marks=pytest.mark.xfail),
        pytest.param("a", "b", "", "", "", marks=pytest.mark.xfail),
    ],
)
def test_databricks_credentials(username, password, host, token, cluster_id):
    databricks = DatabricksCredentials(
        username=username,
        password=password,
        host=host,
        token=token,
        cluster_id=cluster_id,
    )

    expected_config = {
        "Username": username,
        "Password": password,
        "Host": host,
        "Token": token,
        "Cluster": cluster_id,
    }

    assert databricks.type() == "databricks"
    assert databricks.config() == expected_config


@pytest.mark.parametrize(
    "aws_access_key_id,aws_secret_access_key,emr_cluster_id,emr_cluster_region",
    [
        ("a", "b", "c", "d"),
        pytest.param("", "", "a", "b", marks=pytest.mark.xfail),
        pytest.param("", "", "", "", marks=pytest.mark.xfail),
    ],
)
def test_emr_credentials(
    aws_access_key_id, aws_secret_access_key, emr_cluster_id, emr_cluster_region
):
    emr = EMRCredentials(
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        emr_cluster_id=emr_cluster_id,
        emr_cluster_region=emr_cluster_region,
    )

    expected_config = {
        "AWSAccessKeyId": aws_access_key_id,
        "AWSSecretKey": aws_secret_access_key,
        "ClusterName": emr_cluster_id,
        "ClusterRegion": emr_cluster_region,
    }

    assert emr.type() == "emr"
    assert emr.config() == expected_config
